// *******************************************************************
// 
//  A Teaching Environement for Remote Lecturing in Higher Education
//  
//         Prototype Implementation of a Digital Lecture Board
//
//                          Werner Geyer
//                      University of Mannheim
//                     Praktische Informatik IV
//                              L 15, 16
//	                    68131 Mannheim
//                             Germany
//           
//              geyer@pi4.informatik.uni-mannheim.de
//              
// *******************************************************************
//
//  Copyright (c) 1996, 1997, 1998
//     
//  by Lehrstuhl Praktische Informatik IV
//     University of Mannheim
//     Germany
//     
//  All rights reserved.
//              
//  Permission to use, copy, modify, and distribute this
//  software and its documentation for any purpose and without
//  fee is hereby granted, provided that the above copyright
//  notice appear in all copies and that both that copyright
//  notice and this permission notice appear in supporting
//  documentation.
//  
//  We make no representations about the suitability of this 
//  software for any purpose.  It is provided "as is" without 
//  express or implied warranty.
//  
//
//  Author: Werner Geyer
//          Praktische Informatik IV
//          University of Mannheim
//          email: geyer@pi4.informatik.uni-mannheim.de
// 
// *******************************************************************
// program : DADU.h
// authors : Werner Geyer
// date    : 01.05.98
// desc    : definition of DADU class: stores dadu requests; for each
//           dadu request a givethrough timer is set; the dadu request
//           object is deleted after the timer expired OR 
//           whenever a delayed adu response for this request arrived
//           from another member; deletes also corresponding data. when
//           DADU instance is deleted
// *******************************************************************
#ifndef _DADU_H
#define _DADU_H

#define GT_TIMER_EXPIRED 1
#define GT_TIMER_RUNNING 0

namespace SMP {
class DADU {
 public:

  DADU() {};
  DADU(unsigned int _ssrc, unsigned int _sn, void * _data, int _size, 
       unsigned int ip) :
    ssrc(_ssrc), sn(_sn), data((char*)_data), size(_size), ip_addr(ip) {
    id = ++id_count; reply_time = 0;}
  DADU(unsigned int _ssrc, unsigned int _sn) :
    ssrc(_ssrc), sn(_sn), data(NULL), size(0) {}

  static unsigned int id_count;
  unsigned int id;
  unsigned int ssrc;
  unsigned int sn;
  unsigned int ip_addr;
  unsigned int reply_time;
  int status;
  char * data;
  int size;

  bool operator<(const DADU& dadu) {
    return((ssrc < dadu.ssrc) && (sn < dadu.sn));}
  bool operator!=(const DADU& dadu) {
    return((ssrc != dadu.ssrc) || (sn != dadu.sn));}
  bool operator==(const DADU& dadu) {
    return((ssrc == dadu.ssrc) && (sn == dadu.sn));}
};
}

#endif
