#include "Demon.h"
#include "iohelpers.h"
#include "SMP.h"

#ifdef WIN32
using namespace std;
#endif

#ifdef _SMP_ERROR
#include <bitset>
#endif

/* global variables */
SMP::TimerList SMP::Demon::timerlist;
//SORTLIST<Timer> Demon::timerlist;
SMP::Timer* SMP::Demon::Current;
unsigned int SMP::DADU::id_count;
SMP::Demon* SMP::Demon::Deamon = NULL;

//#define _SMP_SENDDELAYED


#ifdef _SMP_DATATRACE
int c_count = 0;
#endif

#define xHANDLE_SIGKILL


extern "C" {
  void linksocket(int fd, int mask, Tcl_FileProc * callback);
  void unlinksocket(int fd);
  int rx_sock_init(char * dhost, int videoserverport, int t_flag);
  int tx_sock_init(char * dhost, int videoserverport, int t_flag);

#ifdef WIN32
  void clearReadSocketMessageQueue(int fd);
  void restoreReadSocketMessageQueue(int fd);
#endif
};

/*********************************************************************
 Name    : Client::Client
 Param   :
 Result  :
 Desc    : Constructor of class Client, used by class Demon
 Calledby:
 Uses    :
 *********************************************************************/
SMP::Client::Client() {
  actNumber = 0;
  for (int x = 0; x < MAX_CLIENTS; x++) {
    seqNumber[x] = 0;
    belegt[x] = false;
    loggedOn[x] = false;
    MaxKbps[x] = 0;
    dadu_seqno[x] = 0;
  }
}

/*********************************************************************
 Name    : Client::~Client
 Param   :
 Result  :
 Desc    : Destructor of class Client
 Calledby:
 Uses    :
 *********************************************************************/
SMP::Client::~Client() {
  for (int x = 0; x < MAX_CLIENTS; x++)
    if ( (belegt[x] == true) && (loggedOn[x] == true) ) {
    }
}

/*********************************************************************
 Name    : Client::search_pos
 Param   :
 Result  : index of free docking station for Client (Application)
           -1 if no free position
 Desc    :
 Calledby: Demon:run
 Uses    :
 *********************************************************************/
int SMP::Client::search_pos() {
  if (actNumber >= MAX_CLIENTS) return -1;

  int index = 0;
  while (belegt[index] == true)
    index++;

  belegt[index] = true;
  actNumber++;

  return index;
}



/*********************************************************************
 Name    : Client::getNetsockPosition
 Param   : fd: file descriptor
 Result  : toHandle
 Desc    :
 Calledby: Demon::HandleIncomingData
 Uses    :
 *********************************************************************/
int SMP::Client::getNetsockPosition(int fd) {
  for (int i=0; i < MAX_CLIENTS; i++) {
    if (msock_r[i] == fd) {
      return i;
    }
  }

  return -1;
}


/*********************************************************************
 Name:      Client::FindDADU()
 Param:     toHandle and id of DADU
 Result:    pointer to ADU on success otherwise NULL
 Desc:      finds DADU request with specified id
 *********************************************************************/
SMP::DADU * SMP::Client::FindDADU(int toHandle, unsigned int id) {

  DADU * dadu = dadulist[toHandle].First();
  while(dadu) {
    if (dadu->id == id) return(dadu);
    dadu = dadulist[toHandle].Next();
  }
  return(NULL);
}

/*********************************************************************
 Name    : Demon::Demon
 Param   :
 Result  :
 Desc    : Constructor of class Demon
 Calledby:
 Uses    :
 *********************************************************************/
SMP::Demon::Demon()
  : CHECKPERIOD(60) {

#ifdef HANDLE_SIGKILL
#ifdef WIN32
  signal(SIGABRT, ShutDown);
  signal(SIGILL, ShutDown);
  signal(SIGTERM, ShutDown);
  signal(SIGSEGV, ShutDown);
  signal(SIGINT, ShutDown);
#else
  signal(SIGKILL, ShutDown);
  signal(SIGINT, ShutDown);
  signal(SIGQUIT, ShutDown);
  signal(SIGTERM, ShutDown);
  signal(SIGPIPE, SIG_IGN);
#endif
#endif

  // *******************************************************************
  // reset global id counter for delayed adus
  // *******************************************************************
  DADU::id_count = 0;


#ifdef WIN32
  // initialize random generator
  srand( (unsigned)time( NULL ) );

  WSADATA WSAdata;

  /* XXX
   * initialize our socket interface plus the tcl 7.5 socket
   * interface (since they redefine some routines we call).
   * eventually we should just call the tcl sockets but at
   * the moment that's hard to set up since they only support
   * tcp in the notifier.
   */

  if (WSAStartup(MAKEWORD(2, 0), &WSAdata)) {
    fprintf(SMP::debug, "Demon::Demon - Windows Sockets init failed $%u\n",  WSAGetLastError());
    abort();
  }

  /*
   * Increase the application queue size from default value of 8.
   * At the default value, cross application SendMessage of WM_KILLFOCUS
   * will fail because the handler will not be able to do a PostMessage!
   * This is only needed for Windows 3.x, since NT dynamically expands
   * the queue.
   */

  SetMessageQueue(64);

#endif
  char localHost[64];
  char hostInAddr[32];
  gethostname(localHost, sizeof(localHost));
  struct hostent* hp = gethostbyname(localHost);
  struct in_addr* ptr = (struct in_addr *) hp->h_addr_list[0];
  strcpy(hostInAddr,inet_ntoa(*ptr));
  HostIpAddr = inet_addr(hostInAddr);

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "\n%s\n", hostInAddr);
#endif

  // ---------- Set Rate Control Timer -------------------------------------
  //
  // currently we use one rate control timer for all clients
  // the timer is called every second and checks credits of
  // all clients, we use the handle -1 for this global timer
  // (traditionaly only client specific timers have been supported
  // so this is a little bit quick and dirty
  SetTimer(-1, 0, 0, Timer::RATECTL,
           0, NULL, 0, 0, false, 0, false, FIRSTTIME);

  last_call = GetTimestamp();
}

/*********************************************************************
 Name    : Demon::ShutDown
 Param   :
 Result  :
 Desc    : free used objects and shut down SMP process
 Calledby:
 Uses    :
 *********************************************************************/
void SMP::Demon::ShutDown(int) {

  //fprintf(SMP::debug, "shut down SMP\n");
  Client* cp = &Deamon->c;

  for (int x = 0; x < MAX_CLIENTS; x++) {
    if ( (cp->belegt[x] == true) && (cp->loggedOn[x] == true) ) {
      shutdown(cp->msock_w[x], 2);
      closesocket(cp->msock_w[x]);
      shutdown(cp->msock_r[x], 2);
      closesocket(cp->msock_r[x]);
    }
  }

#ifdef WIN32
  WSACleanup();
#endif
}



/*********************************************************************
 Name    : Demon::logOn
 Param   :
 Result  :
 Desc    : insert new client/application
 Calledby: Demon::action
 Uses    :
 *********************************************************************/
void SMP::Demon::logOn(int toHandle, LogOnInfo* logOnInfo) {

#ifdef _SMP_SESS
  fprintf(SMP::debug, "Demon::logOn(%u)\n", toHandle);
#endif

  if (!Deamon->c.accepted[toHandle]) return;

  /* check internet address */
  if (is_multicast(logOnInfo->multicastAddr))
    c.Multicast[toHandle] = true;
  else
    c.Multicast[toHandle] = false;

  // *******************************************************************
  // open send and receive sockets
  // *******************************************************************
  if (OpenReceiveSocket(toHandle, logOnInfo) == -1) {
    fprintf(SMP::debug, "Demon::logOn() : could not open receive socket\n");
    return;
  }


  linksocket(c.msock_r[toHandle], TCL_READABLE, &HandleIncomingData);

  //fprintf(SMP::debug, "Demon::logon - client %u receive socket %u\n", toHandle, c.msock_r[toHandle]);

  if (OpenSendSocket(toHandle, logOnInfo) == -1) {
    fprintf(SMP::debug, "Demon::logOn() : could not open send socket\n");
    return;
  }

  // *******************************************************************
  // read logOnInfo values
  // *******************************************************************
  c.SourceOrdering[toHandle] = logOnInfo->SourceOrdering;
  c.LateJoin[toHandle] = logOnInfo->LateJoin;
  c.SharedData[toHandle] = logOnInfo->SharedData;
  //c.Loopback[toHandle] = logOnInfo->Loopback;
  c.GlobalTTL[toHandle] = logOnInfo->ttl;
  c.SessionBandwidth[toHandle] = logOnInfo->Bandwidth / 10;

  switch (logOnInfo->ServiceClass) {
  case 'A':
    c.ServiceClassFactor[toHandle] = 0.11;
    break;
  case 'B':
    c.ServiceClassFactor[toHandle] = 0.08;
    break;
  default : // 'C'
    c.ServiceClassFactor[toHandle] = 0.05;
    break;
  }

  // *******************************************************************
  // set initial values
  // *******************************************************************
  c.Credit[toHandle] = c.SessionBandwidth[toHandle];
  c.OpenRequest[toHandle] = false;
  c.DADURequest_ReceiveBuffer[toHandle].BufferFree = true;
  c.DADURequest_SendBuffer[toHandle].BufferFree = true;
  c.FirstLateJoinInd[toHandle] = true;

  // *******************************************************************
  // Insert Application as Member in Group List
  // *******************************************************************
  unsigned int ownSSRC = GetSSRCIdentifier();

#ifdef _SMP_ACK
  fprintf(SMP::debug, "Demon::LogOn(%u) : own ssrc = %u\n", toHandle, ownSSRC);
#endif

  Member* m = new Member(ownSSRC, HostIpAddr, 0);
  if (m == NULL) {
    fprintf(SMP::debug, "Demon::logOn - operation new Member() failed socket is being closed - try to connect again\n");
    logOff(toHandle);
    Demon::close(toHandle);
    return;
  }
  c.grouplist[toHandle].Insert(m);
  c.Members[toHandle] = 1;
  c.We[toHandle] = m;
  m->ActiveMember = true;

#ifdef _SMP_DEBUG
  c.grouplist[toHandle].Print();
#endif

  // *******************************************************************
  // Create Member Set
  // *******************************************************************
  c.MSet[toHandle] = new MemberSet;

  // *******************************************************************
  // Initializing State of Client
  // *******************************************************************
  c.loggedOn[toHandle] = true;
  c.seqNumber[toHandle] = 0;
  c.SMCount[toHandle] = 0;
  c.TimeOfLastCheck[toHandle] = GetTimestamp();

  // *******************************************************************
  // say hello to group
  // *******************************************************************
  if (SendJoinMsg(toHandle) < 0) {
    fprintf(SMP::debug, "Demon::logOn() : could not send join message\n");
  }

  // *******************************************************************
  // set first session message timer
  // *******************************************************************
  double Laufzeit = rtcp_intervall(1, c.ServiceClassFactor[toHandle]*
				   c.SessionBandwidth[toHandle],
				   sizeof(SMPHeader) + 28 +5*4,
				   &(c.AvgSMSize[toHandle]), true);
#ifdef _SMP_TIMER
  fprintf(SMP::debug, "RTCP Interval: %u\n", Laufzeit);
#endif

#ifdef _SMP_SESS
  fprintf(SMP::debug, "Demon::logOn(%u) : set first session timer\n", toHandle);
#endif

  SetTimer(toHandle, c.We[toHandle]->SSRC, 0, Timer::SESSMESS,
	   0, m, c.We[toHandle]->SSRC, Laufzeit, false, 0, 0, FIRSTTIME);
}

/*********************************************************************
 Name    : Demon::logOff
 Param   :
 Result  :
 Desc    : ends membership of application/client
 Calledby: Demon::action
 Uses    :
 *********************************************************************/
void SMP::Demon::logOff(int toLogOff) {

  if (!c.We[toLogOff]) return;
  if (!c.loggedOn[toLogOff]) return;

#ifdef _SMP_SESS
  fprintf(SMP::debug, "Demon::logOff(%u)\n", toLogOff);
#endif

  // *******************************************************************
  // say bye to group
  // *******************************************************************
  if (SendLeaveMsg(toLogOff) < 0) {
    fprintf(SMP::debug, "Demon::LogOff() : could not send leave message\n");
  }

  // *******************************************************************
  // delete all timers
  // *******************************************************************
  DeleteAllTimerForLeavingApplication(toLogOff, c.We[toLogOff]->SSRC);

  c.grouplist[toLogOff].Destroy();
  c.CollisionList[toLogOff].clear();
  c.ratecontrolqueue[toLogOff].Destroy();

  c.OpenRequest[toLogOff] = false; //lj
  c.MaxKbps[toLogOff] = 0;
  c.SessionBandwidth[toLogOff] = 0;
  c.Credit[toLogOff] = 0;
  c.loggedOn[toLogOff] = false;


  unlinksocket(c.msock_r[toLogOff]);

  shutdown(c.msock_r[toLogOff], 2);
  closesocket(c.msock_r[toLogOff]);
  shutdown(c.msock_w[toLogOff], 2);
  closesocket(c.msock_w[toLogOff]);

  //Tcl_DeleteFileHandler(c.msock_r[toLogOff]);

  delete c.MSet[toLogOff];
}

/*********************************************************************
 Name    : Demon::sendCtlMsg
 Param   :
 Result  :
 Desc    : send Control Message to group
 Calledby: Demon::logOn, Demon::logOff, Demon::SSRCCollision
           Demon::ExecuteTimerAction
 *********************************************************************/
bool SMP::Demon::sendCtlMsg(int toHandle, void* msg, int size, bool block) {

  if (!c.We[toHandle]) return false;
  if (!c.loggedOn[toHandle]) return false;

  if (::send(c.msock_w[toHandle], (char*)msg, size, 0) < 0) {
#ifdef WIN32
    if ((WSAGetLastError() == WSAECONNREFUSED) && !c.Multicast[toHandle]) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::sendCtlMsg - function send cannot access multicast socket %u socket is being closed\n",
	      WSAGetLastError());
      logOff(toHandle);
      Demon::close(toHandle);
      return false;
    }
#else
    //UNIX
    /* we are trying to send unicast data with no recipants => reonnect */
    if ((errno == ECONNREFUSED) && (!c.Multicast[toHandle])) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::sendCtlMsg - function send cannot access multicast socket\nsocket is being closed - try to connect again\n");
      logOff(toHandle);
      Demon::close(toHandle);
      return false;
    }
#endif
  }
  return true;
}

/*********************************************************************
 Name    : Demon::close
 Param   :
 Result  :
 Desc    : closes connection (socket) to leaving client/application
 Calledby: Demon::action
 *********************************************************************/
void SMP::Demon::close(int toDie) {

  if (c.loggedOn[toDie]) {
    logOff(toDie);
  }
  IndClose(toDie, 0);
  c.belegt[toDie] = false;
  c.accepted[toDie] = false;     /* reset accepted */
  c.actNumber--;
}

/*********************************************************************
 Name    : Demon::GetSequenceNumber
 Param   : client to handle
 Result  : new seq.no.
 *********************************************************************/
unsigned int SMP::Demon::GetSequenceNumber(int toHandle) {
  return(++(c.seqNumber[toHandle]));
}


/*********************************************************************
 Name    : Demon::HandleIncomingData
 Param   :
 Result  :
 Desc    : handles incoming net traffic
 Calledby: Demon::run
 Uses    :
 *********************************************************************/
void SMP::Demon::HandleIncomingData(ClientData clientData, int mask) {

  int toHandle = Deamon->c.getNetsockPosition((int)clientData);

  //  if (toHandle == 2)
  //  fprintf(SMP::debug, "Demon::HandleIncomingData - got data for client %u socket %u\n", toHandle, (int)clientData);

#ifdef _AIX
  long unsigned int fromlen = sizeof(sockaddr_in);
#else
  int fromlen = sizeof(sockaddr_in);
#endif

  struct sockaddr_in fromaddr;
  //int lenPacket;
  bool SenderUnKnown = false;
  int msg_size;

  if (!Deamon->c.loggedOn[toHandle]) return;

#ifdef WIN32
  clearReadSocketMessageQueue(Deamon->c.msock_r[toHandle]);
#endif

  // *******************************************************************
  // look up size of received packet
  // *******************************************************************

#ifdef WIN32
  // Windows
  unsigned long pSize = 0;
  if (ioctlsocket(Deamon->c.msock_r[toHandle], FIONREAD, &pSize) != 0) {
    fprintf(SMP::debug, "Demon::HandleIncomingData - ioctlsocket failed: %u socket is being cloesd\n", WSAGetLastError());
    Deamon->logOff(toHandle);
    Deamon->close(toHandle);
    return;
  }
  int packetSize = (int)pSize;
  msg_size = packetSize;

  //fprintf(SMP::debug, "Demon::HandleIncomingData - %u bytes on socket\n", packetSize);
#else
#ifdef linux
  SMPHeader* lookupHeader = new SMPHeader;
  if ((msg_size = recvfrom(Deamon->c.msock_r[toHandle],
			   (char*)lookupHeader,
			   sizeof(SMPHeader),
			   MSG_PEEK,
			   (sockaddr *)&fromaddr,
			   (socklen_t*)&fromlen)) < 0) {
    fprintf(SMP::debug, "Demon::HandleIncomingData - function recvfrom cannot access socket (MSG_PEEK)\nsocket is being closed\n");
    Deamon->logOff(toHandle);
    Deamon->close(toHandle);
    return;
  }

  int packetSize = ntohs(lookupHeader->SZ);
  delete lookupHeader;
#else
  // UNIX
  SMPHeader* lookupHeader = new SMPHeader;
  if ((msg_size = recvfrom(Deamon->c.msock_r[toHandle],
			   (char*)lookupHeader,
			   sizeof(SMPHeader),
			   MSG_PEEK,
			   (sockaddr *)&fromaddr,
			   &fromlen)) < 0) {
    fprintf(SMP::debug, "Demon::HandleIncomingData - function recvfrom cannot access socket (MSG_PEEK)\nsocket is being closed\n");
    Deamon->logOff(toHandle);
    Deamon->close(toHandle);
    return;
  }

  int packetSize = ntohs(lookupHeader->SZ);
  delete lookupHeader;
#endif
#endif


  // *******************************************************************
  // allocate memory for packet
  // *******************************************************************
  char * data;
  if ((data = Deamon->AllocatePacketBuffer(toHandle, packetSize)) == NULL) {
    fprintf(SMP::debug, "Demon::HandleIncomingData(%u) : could not allocate packet buffer ... closing socket\n", toHandle);
    return;
  }


  // *******************************************************************
  // read packet into memory
  // *******************************************************************
  int bread;
#ifdef linux
  if ((bread = recvfrom(Deamon->c.msock_r[toHandle], data, packetSize, 0,
			(sockaddr *)&fromaddr, (socklen_t*)&fromlen)) < 0) {
    fprintf(SMP::debug, "Demon::HandleIncomingData - function recvfrom cannot access socket (READ)\nsocket is being closed - try to connect again\n");
    Deamon->FreePacketBuffer(toHandle, data, bread);
    Deamon->logOff(toHandle);
    Deamon->close(toHandle);
    return;
  }
#else
  if ((bread = recvfrom(Deamon->c.msock_r[toHandle], data, packetSize, 0,
			(sockaddr *)&fromaddr, &fromlen)) < 0) {
    fprintf(SMP::debug, "Demon::HandleIncomingData - function recvfrom cannot access socket (READ)\n");
#ifdef WIN32
    fprintf(SMP::debug, "%u\n", WSAGetLastError());
#endif
    fprintf(SMP::debug, "socket is being closed - try to connect again\n");
    Deamon->FreePacketBuffer(toHandle, data, bread);
    Deamon->logOff(toHandle);
    Deamon->close(toHandle);
#ifdef WIN32
    restoreReadSocketMessageQueue(Deamon->c.msock_r[toHandle]);
#endif
    return;
  }
#endif


#ifdef WIN32
  restoreReadSocketMessageQueue(Deamon->c.msock_r[toHandle]);
#endif

  // *******************************************************************
  // read payload type
  // *******************************************************************
  SMPHeader* header = (SMPHeader*)data;

  int receivedPT = ntohs(header->FLAGS)&SMP_PT;

#ifdef _SMP_HEADER
  unsigned int receivedSSRC = ntohl(header->SSRC);
  unsigned int receivedTS = ntohl(header->TS);
  unsigned int receivedSN = ntohl(header->SN);
  bool receivedX = ntohs(header->FLAGS)&SMP_X;
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

  fprintf(SMP::debug, "\t   packet from (%u,%u,%u,%u): X = %u PT = %u SN = %u Size = %u\n", receivedSSRC,
	  receivedSN, receivedTS, receivedPT, (ntohs(header->FLAGS)&SMP_X), (ntohs(header->FLAGS)&SMP_PT),
	  ntohl(header->SN), packetSize);
#endif

  // *******************************************************************
  // check payload
  // *******************************************************************
  if (receivedPT == SMPHeader::DATA) {
    if (Deamon->ReceiveData(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive data error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::DATA_UNRELIABLE) {
    if (Deamon->ReceiveUnreliableData(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive data error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::JOIN) {
    if (Deamon->ReceiveJoin(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive join error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::LEAV) {
    if (Deamon->ReceiveLeave(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive leave error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::REQ) {
    if (Deamon->ReceiveRequest(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) receive request error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::SMES) {
    if (Deamon->ReceiveSessionMsg(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive sessmsg error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::DADURQ) {
    if (Deamon->ReceiveDADUReq(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive DADU req error\n", toHandle);
    }
    return;
  }
  if (receivedPT == SMPHeader::DADURP) {
    if (Deamon->ReceiveDADURsp(toHandle, data, bread, fromaddr) < 0) {
      fprintf(SMP::debug, "Demon::receive(%u) : receive DADU rsp error\n", toHandle);
    }
    return;
  }

  if (msg_size > RECV_BUF) {
    fprintf(SMP::debug, "Demon::receive() : datagram to large\n");
    msg_size = RECV_BUF;
  }

  /* data is unknown => clean up buffer */
  if (recv(Deamon->c.msock_r[toHandle], Deamon->recv_buf, msg_size, 0) < 0) {
    fprintf(SMP::debug, "Demon::receive() : could not receive unknown data type\n");
  }
}


/*********************************************************************
 Name    : Demon::ComputeSessionMessage
 Param   :
 Result  :
 Desc    : analyses received session message
 Calledby: Demon::receive
 Uses    :
 *********************************************************************/
void SMP::Demon::ComputeSessionMessage(int toHandle, char* Header,
				  unsigned int receivedSSRC,
                                  unsigned int receivedTS,
                                  Member* MessageSender) {

  SMPSessionMessageElement* Element = (SMPSessionMessageElement *)
    (Header + (sizeof(SMPHeader) +
	       sizeof(SMPSessionMessageCount)));
  SMPSessionMessageCount* Count = (SMPSessionMessageCount *)
    (Header + sizeof(SMPHeader));

#ifdef _SMP_SESS
  fprintf(SMP::debug, "Demon::ComputeSessionMessage - nbr of Members: %u\n", Count->i);
#endif

  // *******************************************************************
  // browse through all elements
  // *******************************************************************
  for (int i = 0; i < ntohl(Count->i); i++) {

#ifdef _SMP_SESS
    fprintf(SMP::debug, "[%u]: %u, %u, %u\n", i, ntohl(Element->SSRC),
	    ntohl(Element->SN), ntohl(Element->HSN));
#endif

    Member m(ntohl(Element->SSRC));
    Member* MentionedMember = c.grouplist[toHandle].Find(&m);

    // *******************************************************************
    // member not found => create
    // *******************************************************************
    if (MentionedMember == NULL) {

      MentionedMember = new Member(ntohl(Element->SSRC), 0, 0);
      if (!MentionedMember) {
	fprintf(SMP::debug, "Demon::ComputeSessionMessage - operation new Member failed, close socket\n");
	logOff(toHandle);
	Demon::close(toHandle);
	return;
      }
      // SN und HighestNum im Konstruktor zu null gesetzt

      MentionedMember->timestampSent = receivedTS;
      unsigned int time = GetTimestamp();
      MentionedMember->timestampReceived = time;
      MentionedMember->LastSessionMessage = time;
      MentionedMember->delay = DEFAULTDISTANCE;
      c.MSet[toHandle]->Insert(MentionedMember);
      c.Members[toHandle]++;
      c.grouplist[toHandle].Insert(MentionedMember);

      if (ntohs(Element->Active)) {
	MentionedMember->ActiveMember = true;
      }
      else {
	// bisher unbekanntes, aber bereits ausgeschiedenes Mitglied
	MentionedMember->ActiveMember = false;
      }

      if (!c.LateJoin[toHandle]) {
	/*
	   if (c.FirstLateJoinInd[toHandle]) {
	   #ifdef _SMP_LJ
	   fprintf(SMP::debug, "Demon::ComputeSessionMessage() : send LJIND to app\n");
	   #endif
	   LateJoinInd(toHandle, ntohl(Element->SSRC));
	   c.FirstLateJoinInd[toHandle] = false;
	   }
	*/

	// regulaeren Repair-Mechanismus unterdruecken
	MentionedMember->HighestNum = ntohl(Element->SN);
      } else {

	for (unsigned int x = 1; x <= ntohl(Element->SN); x++) {

#ifdef _SMP_REQTIMER
	  fprintf(SMP::debug, "Demon::ComputeSessionMsg() : new request timer for %u toHandle = %u delay = %u\n",
		  x, toHandle, MentionedMember->delay);
#endif
	  // Timer stellen fuer Packet x
	  Timer* forTimer = FindTimer(toHandle, Timer::REQUEST,
				      ntohl(Element->SSRC), x);
	  if (forTimer == NULL) {
	    SetTimer(toHandle, ntohl(Element->SSRC),
		     x, Timer::REQUEST, MentionedMember->delay,
		     MentionedMember, c.We[toHandle]->SSRC,
		     0, true, 0, 0, FIRSTTIME);
	  }

	}
      }
      MentionedMember->SN = ntohl(Element->SN);

    }

    // *******************************************************************
    // member known =>
    // - check active state
    // - check received packets
    // - update op list
    // *******************************************************************
    else {

      if ((ntohl(Element->Active) == 0) && (MentionedMember->ActiveMember == true) ) {
	MentionedMember->ActiveMember = false;

	if (!c.LateJoin[toHandle]) {
	  UpdateAllOpLists(toHandle, MentionedMember->SSRC, 0);
	}
      }
      // *******************************************************************
      // everything is ok => update op list (ack)
      // *******************************************************************
      else {
	if (!c.LateJoin[toHandle]) {
	  UpdateOpList(toHandle, MentionedMember, receivedSSRC, ntohl(Element->HSN));
	}
      }

      // *******************************************************************
      // lost packets ? => set request timer
      // *******************************************************************
      if (MentionedMember->SN < ntohl(Element->SN)) {
	unsigned int ExpectedSequenceNumber = MentionedMember->SN + 1;
	unsigned int ReceivedSequenceNumber = ntohl(Element->SN);

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Demon::ComputeSessionMessage(%u) : REQUESTING PACKETS\n%u to %u from %u we = %u\n",
		toHandle, ExpectedSequenceNumber, ReceivedSequenceNumber, MentionedMember->SSRC, c.We[toHandle]->SSRC);
#endif

	for (unsigned int j = ExpectedSequenceNumber; j <= ReceivedSequenceNumber; j++) {

#ifdef _SMP_REQTIMER
	  fprintf(SMP::debug, "Demon::ComputeSessionMsg() : new request timer for %u toHandle = %u delay = %u\n",
		  j, toHandle, MentionedMember->delay);
#endif
	  Timer* forTimer = FindTimer(toHandle, Timer::REQUEST,
				      MentionedMember->SSRC, j);
	  if (forTimer == NULL) {

	    SetTimer(toHandle, MentionedMember->SSRC,
		     j, Timer::REQUEST,MentionedMember->delay,
		     MentionedMember, c.We[toHandle]->SSRC, 0,
		     true, 0, false, FIRSTTIME);
	  }

	}
	MentionedMember->SN = ReceivedSequenceNumber;
      }


      // *******************************************************************
      // if our own entry => recalculate distance
      // *******************************************************************
      if (MentionedMember->SSRC == c.We[toHandle]->SSRC) {

#ifdef _SMP_DELAY
	fprintf(SMP::debug, ">> vorher delay(%u) = %u\n", toHandle, MessageSender->delay);
#endif

	MessageSender->delay =
	  (int)(.5 + (ALPHA*MessageSender->delay) + (1-ALPHA)
		*GetDistanceInMillis(ntohl(Element->Timestamp1),
				     ntohl(Element->Delta)));

#ifdef _SMP_DELAY
	fprintf(SMP::debug, ">> nachher delay(%u) = \n", toHandle, MessageSender->delay);
#endif

      }
    }
    Element += 1;
  }
}



/*********************************************************************
 Name    : Demon::IndData
 Param   :
 Result  :
 Desc    : informs application/client about a received ADU
 Calledby: Demon::receive
 Uses    :
 *********************************************************************/
void SMP::Demon::IndData(int toHandle, char* data, int size, unsigned int senderIPAddr) {

#ifdef _SMP_DATATRACE
  fprintf(SMP::debug, "Demon::IndData(%u) : Packet with sn %u given to app\n",
	  toHandle, ntohl(((SMPHeader*)data)->SN));
#endif

#ifdef _SMP_SORDER
  fprintf(SMP::debug, "Client %u Packet given to application", toHandle);
#endif

  int _size = size - sizeof(SMPHeader);

  Indication ind;
  ind.buf  = new Buffer(_size);
  ind.ServiceType = Indication::DATAIND;
  ind.senderIPAddr = senderIPAddr;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  memcpy(ind.buf->getCont(), data + sizeof(SMPHeader), _size);

  SMP::smp_pointer->group[toHandle].Callback(ind);
}


/*********************************************************************
 Name    : Demon::IndUnreliableData
 Param   :
 Result  :
 Desc    : informs application/client about a received unreliable ADU
 Calledby: Demon::receive
 Uses    :
 *********************************************************************/
void SMP::Demon::IndUnreliableData(int toHandle, char* data, int size, unsigned int senderIPAddr) {

#ifdef _SMP_DATATRACE
  fprintf(SMP::debug, "Demon::IndUnreliableData(%u) packet given to app\n", toHandle);
#endif

#ifdef _SMP_SORDER
  fprintf(SMP::debug, "Client %u Packet given to application\n", toHandle);
#endif

  int _size = size - sizeof(SMPHeader);

  Indication ind;
  ind.buf  = new Buffer(_size);
  ind.ServiceType = Indication::UNRLDATAIND;
  ind.senderIPAddr = senderIPAddr;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  memcpy(ind.buf->getCont(), data + sizeof(SMPHeader), _size);

  SMP::smp_pointer->group[toHandle].Callback(ind);
}

/*********************************************************************
 Name    : Demon::IndClose
 Param   : client toHandle, client SSRC
 Desc    : informs application/client about closed communication channel
 Calledby: Demon::logOff
 *********************************************************************/
void SMP::Demon::IndClose(int toHandle, unsigned int SSRC) {
#ifdef _SMP_DATATRACE
  fprintf(SMP::debug, "Demon::IndClose()\n");
#endif

  Indication ind;
  ind.ServiceType = Indication::CLOSEIND;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  SMP::smp_pointer->group[toHandle].Callback(ind);
}


/*********************************************************************
 Name    : Demon::IndDADUReqFailed
 Param   : client toHandle
 Desc    : informs application/client that a sended delayed adu request
           was not answered by a group member
 Calledby: Demon::ExecuteTimerAction
 *********************************************************************/
void SMP::Demon::IndDADUReqFailed(int toHandle) {
#ifdef _SMP_DATATRACE
  fprintf(SMP::debug, "Demon::IndDADUReqFailed()");
#endif

  Indication ind;
  ind.ServiceType = Indication::DADUFAIL;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  SMP::smp_pointer->group[toHandle].Callback(ind);
}


/*********************************************************************
 Name    : Demon::IndLateJoin
 Param   :
 Result  :
 Desc    : informs application/client about the necessity for a smart
           late join
 Calledby: Demon::receive, Demon::ComputeSessionMessage
 Uses    :
 *********************************************************************/
void SMP::Demon::IndLateJoin(int toHandle, unsigned int SSRC) {
#ifdef _SMP_LJ
  fprintf(SMP::debug, "Demon::IndLateJoin() : sending late join ind to app");
#endif

  Indication ind;
  ind.ServiceType = Indication::LJNEEDED;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  SMP::smp_pointer->group[toHandle].Callback(ind);
}

/*********************************************************************
 Name    : Demon::GetDistanceInMillis
 Param   :
 Result  :
 Desc    :
 Calledby: Demon::ComputeSessionMessage
 Uses    :
 *********************************************************************/
unsigned int SMP::Demon::GetDistanceInMillis(unsigned int T1, unsigned int Delta) {
  unsigned int Now = GetTimestamp();
  if ( (Now - T1 - Delta) < 0) {

#ifdef _SMP_DELAY
    fprintf(SMP::debug, "Demon::GetDistanceInMillis() : overflow, using default distance = %u\n", DEFAULTDISTANCE);
#endif

    return DEFAULTDISTANCE;
  }

#ifdef _SMP_DELAY
  fprintf(SMP::debug, "Demon::GetDistanceInMillis() = %u now = %u t1 = %u delta = %u\n",
	  ((Now - T1 - Delta)/2), Now, T1, Delta);
#endif

  return (Now - T1 - Delta)/2;
}


/*********************************************************************
 Name    : Demon::GetTimestamp
 Param   :
 Result  :
 Desc    :
 Calledby: Demon::logOn
           Demon::logOff
           Demon::receive
           Demon::ComputeSessionMessage
           Demon::AddSessionMessageBody
           Demon::GetDistanceInMillis
           Demon::SSRCCollision
           Demon::ExecuteTimerAction
           Demon::CheckLivenessOfMembers
 Uses    :
 *********************************************************************/
unsigned int SMP::Demon::GetTimestamp() {
#ifdef WIN32

   struct _timeb timebuffer;
   _ftime(&timebuffer);
	 return (1000*timebuffer.time + timebuffer.millitm);
#else
  struct timeval tp;
  gettimeofday(&tp, 0);
  return ((tp.tv_sec * 1000) + (tp.tv_usec / 1000));
#endif
}


/*********************************************************************
 Name    : Demon::GetSSRCIdentifier
 Param   :
 Result  :
 Desc    :
 Calledby: Demon::logOn
           Demon::SSRCCollision
 Uses    :
 *********************************************************************/
unsigned int SMP::Demon::GetSSRCIdentifier () {
#ifdef WIN32
   struct _timeb timebuffer;
   _ftime(&timebuffer);
	 return (1000*timebuffer.time + timebuffer.millitm);
#else
	struct timeval tp;
  gettimeofday(&tp, 0);
  return ((tp.tv_sec * 1000) + (tp.tv_usec / 1000));
#endif

}


/*********************************************************************
 Name    : Demon::SSRCCollision
 Param   :
 Result  :
 Desc    : handles collision between SSRC`s of group members
 Calledby: Demon::receive
 Uses    :
 *********************************************************************/
void SMP::Demon::SSRCCollision(int toHandle, unsigned int CollisionSSRC) {

 if (c.We[toHandle]->SSRC == CollisionSSRC) {
     // Wir selbst sind in die Kollision verwickelt
     // 1. Senden eines Leave
     // 2. SSRC in Kollisionsliste aufnehmen
     // 3. Neues SSRC generieren + SN auf 0 setzen
     // 4. TimerListe mit neuem SSRC updaten
     // 5. neue SSRC an Anwendung senden

     // ------------- Sending Leave-Message to the entire Group ---------
     {
       SMPHeader Header;
       u_int16 flags = 0 << 15 | SMPHeader::LEAV;
       Header.FLAGS = htons(flags);
       Header.SSRC = htonl(c.We[toHandle]->SSRC);
       Header.TS   = htonl(GetTimestamp());
       Header.SZ   = htons(sizeof(SMPHeader));
       sendCtlMsg(toHandle, &Header, sizeof(SMPHeader), false);
     }

     unsigned int OldSSRC = c.We[toHandle]->SSRC;
     unsigned int NewSSRC = GetSSRCIdentifier();
     c.CollisionList[toHandle].insert(OldSSRC);

     c.We[toHandle]->SSRC = NewSSRC;
     c.We[toHandle]->SN = 0;
     UpdateTimerList(OldSSRC, c.We[toHandle]->SSRC);
  }
  else {
     // Fremd-Kollision entdeckt
     // Die kollidierte SSRC kommt in die Leaved-Liste
     // Alle weiteren Packete beider Sourcen werden ignoriert
     c.CollisionList[toHandle].insert(CollisionSSRC);
  }

}


/*********************************************************************
 Name    : Demon::SetTimer
 Param   :
 Result  :
 Desc    : inserts new timer into timer list
 Calledby: Demon::logOn
           Demon::receive
           Demon::ExecuteTimerAction
 Uses    :
 *********************************************************************/
void SMP::Demon::SetTimer(int toH, unsigned int ssrc, unsigned int SeqN,
                     int action, int delay, Member* sender,
                     unsigned int Erzeuger, double Lauf,
                     bool LR, unsigned char TTL, bool FNack,
                     int count, Timer* NewTimer) {

  if (NewTimer == NULL) {
    // SetTimer wurde nicht aus dem SignalHandler aufgerufen
    // Damit ist die Erzeugung eines neuen Timer unproblematisch. Allerdings
    // kann der neue Timer frueher faellig sein als der bereits laufende.
    // Also muss der laufende Timer gestoppt werden und der neue gesetzt
    // werden.

    NewTimer = new Timer(toH, ssrc, SeqN, action, delay,
			 sender, Erzeuger, Lauf, LR, TTL, FNack, count);

#ifdef _SMP_TIMER
    fprintf(SMP::debug, "Demon::SetTimer() : new timer with id %u time = %u\n",
	    NewTimer->ActionId, NewTimer->GesamtLaufZeit);
#endif

    if (NewTimer == NULL) {
      fprintf(SMP::debug, "Demon::SetTimer - operation new Timer failed, close socket\n");
      logOff(toH);
      Demon::close(toH);
      return;
    }

    if (!timerlist.Empty() && (*Current < *NewTimer)) {
      // timer list is not empty and current timer is smaller
      // than the new timer: this means, we just insert the new
      // timer in the list without modifying any running timers

      timerlist.Insert(NewTimer);
    } else {
      // timer list is empty AND/OR current timer is larger than new timer
      // Thus, set a new system timer

      // reset any running timer
      DeleteRunningTimer();
      timerlist.Insert(NewTimer);

      Current = NewTimer;

      // set new timer
      //struct itimerval LaufZeit;
      //LaufZeit.it_interval.tv_sec = 0;
      //LaufZeit.it_interval.tv_usec = 0;
      //LaufZeit.it_value.tv_sec = (long)(NewTimer->GesamtLaufZeit / 1000);
      //LaufZeit.it_value.tv_usec = (long)((NewTimer->GesamtLaufZeit % 1000)
      //				 * 1000);



#ifdef _SMP_TIMER
      fprintf(SMP::debug, "Demon::SetTimer() : next timer event (vor aktuellem): %u\n",
	      NewTimer->GesamtLaufZeit);
#endif

      // if (setitimer(ITIMER_REAL, &LaufZeit, NULL) == -1) {
      //	perror("Demon::SetTimer - function setitimer failed");
      //	if (toH >=0) {
      //	  fprintf(SMP::debug, "socket is being closed - try to connect again\n");
      //	  logOff(toH);
      //	  Demon::close(toH);
      //	  return;
      //	}
      //  }
      timer = Tcl_CreateTimerHandler(NewTimer->GesamtLaufZeit,
				     (Tcl_TimerProc*)&(SignalHandler),
				     NULL);

#ifdef _SMP_TIMER
      fprintf(SMP::debug, "Demon::SeTimer() : total running time %u\n", NewTimer->GesamtLaufZeit);
#endif

    }
  }
  else {
    // Aufruf dieser Funktion aus dem SignalHandler nach einem Signal
    // vom System. Es darf deshalb kein Speicher mit new allokiert werden,
    // da diese Aktion nicht reentrant waere. Der neue Timer wird in die
    // Timer Liste eingefuegt. Der Alarmtimer wird von dort aus neu gestellt

      NewTimer->ReUse(toH, ssrc, SeqN, action, delay, sender,
                      Erzeuger, Lauf, LR, TTL, FNack, count);
      timerlist.Insert(NewTimer);

#ifdef _SMP_TIMER
      if (!(NewTimer->ActionId == Timer::RATECTL)) {
	fprintf(SMP::debug, "Demon::SetTimer() : reuse timer with id %u time %u\n",
		NewTimer->ActionId, NewTimer->GesamtLaufZeit);
      }
#endif
  }

  // check for error
  if (NewTimer->GesamtLaufZeit > 60000)
    fprintf(SMP::debug, "Demon::SetTimer - microseconds larger than 60000\n");
}


/*********************************************************************
 Name    : Demon::SignalHandler
 Param   :
 Result  :
 Desc    : Handles incoming alarm/timer signal
 Calledby: Operating System
 Uses    :
 *********************************************************************/
void SMP::Demon::SignalHandler(...) {
  if (!Current) return;

#ifdef _SMP_RATECTRL
  fprintf(SMP::debug, "entering signalhandler function...\n");
#endif

  Deamon->last_call = Deamon->GetTimestamp();

#ifdef _SMP_TIMER
  printf(SMP::debug, "timer list:\n");
  timerlist.Print();
#endif

  int no_of_cycles = 0;

  int RestTimeToGo;

  while(Current != NULL && no_of_cycles < 5) {
    unsigned int Now = Deamon->GetTimestamp();
    RestTimeToGo =  Current->GesamtLaufZeit - (Now - Current->StartZeitPunkt);

    if (RestTimeToGo < 10) {

      Timer* TimerToRecycle = timerlist.RemoveFirst();
      if (!TimerToRecycle) {
	fprintf(SMP::debug, "Demon::SignalHandler - TimerToRecycle = NULL\n");
	ShutDown(0);
      }

      Deamon->ExecuteTimerAction(TimerToRecycle);
      Current = timerlist.First();
    } else {
      break;
    }
    no_of_cycles++;
  }

  if (Current) {
    //struct itimerval LaufZeit;
    //LaufZeit.it_interval.tv_sec =  0;
    //LaufZeit.it_interval.tv_usec =  0;

    int timer_value;

    if (RestTimeToGo < 0) {
      //LaufZeit.it_value.tv_sec  = 0;
      //LaufZeit.it_value.tv_usec = 20000;
      timer_value = 20;
    } else {
      //LaufZeit.it_value.tv_sec  = (long)(RestTimeToGo / 1000);
      //LaufZeit.it_value.tv_usec  = (long)((RestTimeToGo % 1000) * 1000);
      timer_value = RestTimeToGo;
    }

#ifdef _SMP_TIMER
    if (Current->ActionId != Timer::RATECTL) {
      fprintf(SMP::debug, "next timer event (regulaer): %u type of event = %u\n",
	     timer_value, Current->ActionId);
    }
#endif
    Deamon->timer = Tcl_CreateTimerHandler(timer_value,
					  (Tcl_TimerProc*)&(SignalHandler),
					  NULL);

    //    if (setitimer(ITIMER_REAL, &LaufZeit, NULL) == -1) {
    //       perror("Demon::SignalHandler - function setitimer failed");
    //    }
  }

#ifdef _SMP_RATECTRL
  fprintf(SMP::debug, "leaving signalhandler function");
#endif

}



/*********************************************************************
 Name    : Demon::ExecuteTimerAction
 Param   :
 Result  :
 Desc    : executes action initiated by expired timer
 Calledby: Demon::SignalHandler
 Uses    :
 *********************************************************************/
void SMP::Demon::ExecuteTimerAction(Timer* TimerToRecycle) {

  int toHandle = TimerToRecycle->toHandle;

  // *******************************************************************
  // a handle of larger 0 or equal to zero indicates a client specific
  // timer, if client is not logged on anymore , do not execute timer
  // a handle 0f -1 indicates the rate control timer
  // *******************************************************************
  if ((toHandle >=0)
      && !(c.loggedOn[toHandle])
      && (TimerToRecycle->ActionId != Timer::RATECTL)) return;

  switch (TimerToRecycle->ActionId) {


    // *******************************************************************
    // REPAIR timer
    // *******************************************************************
  case Timer::REPAIR: {

#ifdef _SMP_TIMER
    fprintf(SMP::debug, "REPAIR-Timer ran out");
#endif

    // *******************************************************************
    // find data location in operation list of sender
    // *******************************************************************
    Operation toFind(TimerToRecycle->SN);

    Operation* RepairPacket =
      TimerToRecycle->Sender->operationslist.Find(&toFind);

    if (RepairPacket == NULL) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : could not find data for repair operation\n", toHandle);
      return;
    }

    // *******************************************************************
    // remember repair time
    // *******************************************************************
    RepairPacket->repair_time = GetTimestamp();

    if (SendRepair(toHandle, RepairPacket->data, RepairPacket->size,
		   TimerToRecycle->TTL) < 0) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : could not send repair\n", toHandle);
      return;
    }

#ifdef _SMP_REQTIMER
    fprintf(SMP::debug, "Client %u sends a Repair for Packet[%u]\n",
	    toHandle, TimerToRecycle->SN);
#endif

    delete(TimerToRecycle);
    break;
  }

  // *******************************************************************
  // REQUEST timer
  // *******************************************************************
  case Timer::REQUEST: {

            // 1. SMP-Paket fuer Request-Message anlegen
            // 2. Body hinzufuegen
            // 3. SMP-Paket versenden
            // 4. Neuen Request-Timer stellen

    if (SendRequest(toHandle, TimerToRecycle->SSRC, TimerToRecycle->SN) < 0) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : could not send request\n", toHandle);
      return;
    }

#ifdef _SMP_DEBUG
    fprintf(SMP::debug, "Sending a Request for Packet[%u] org sender %u, i = %u, ms = %u, delay = %u\n",
	    TimerToRecycle->SN, TimerToRecycle->SSRC, TimerToRecycle->i, TimerToRecycle->GesamtLaufZeit,
	    TimerToRecycle->Delay);
#endif

    if ((!c.SourceOrdering[toHandle]) && (TimerToRecycle->i > Demon::MAX_RETRIES)) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u): request for packet %u from %u failed, giving up\n",
	      toHandle, TimerToRecycle->SN, TimerToRecycle->SSRC);
      delete(TimerToRecycle);
    }

#ifdef _SMP_TIMER
    fprintf(SMP::debug, "Request timer = \n");
#endif

    // *******************************************************************
    // get source of this packet
    // *******************************************************************
    Member ToFind(TimerToRecycle->SSRC);
    Member* originSender = c.grouplist[toHandle].Find(&ToFind);

#ifdef _SMP_REQTIMER
    fprintf(SMP::debug, "Set new request timer for %u toHandle = %u delay = %u\n",
	    TimerToRecycle->SN, toHandle, originSender->delay);
#endif

    if (TimerToRecycle->SN <= originSender->HighestNum) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : request for packet which we already have ... delete timer\n", toHandle);
      delete(TimerToRecycle);
    } else {

      Deamon->SetTimer(toHandle, TimerToRecycle->SSRC,
		      TimerToRecycle->SN,
		      Timer::REQUEST, originSender->delay,
		      TimerToRecycle->Sender,
		      Deamon->c.We[toHandle]->SSRC,
		      0, false, 0, false, (TimerToRecycle->i + 1),
		      TimerToRecycle);
    }

    break;
  }

  // *******************************************************************
  // SESSMESS timer
  // *******************************************************************
  case Timer::SESSMESS: {

    int size;
    if (!Deamon->c.loggedOn[toHandle]) return;

    if ((size = SendSessionMsg(toHandle, TimerToRecycle->SSRC,
			       c.We[toHandle]->SN))< 0) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : could not send session message\n", toHandle);
      return;
    }

#ifdef _SMP_DEBUG
    fprintf(SMP::debug, "Client[%u]: send Session Message\n", toHandle);
#endif

    // *******************************************************************
    // claculate session message interval
    // *******************************************************************
    double Laufzeit = rtcp_intervall(c.Members[toHandle],
				     c.ServiceClassFactor[toHandle]*
				     c.SessionBandwidth[toHandle],
				     size + 28, &(c.AvgSMSize[toHandle]),
				     false);

    // *******************************************************************
    // if logged out => don't set new session message timer
    // *******************************************************************
    if (c.loggedOn[toHandle]) {
      CheckLivenessOfMembers(toHandle);

#ifdef _SMP_TIMER
      fprintf(SMP::debug, "Session timer (%u) = \n", toHandle);
#endif

#ifdef _SMP_SESS
      fprintf(SMP::debug, "set new sessmess timer for client %u\n", toHandle);
#endif

      SetTimer(toHandle, TimerToRecycle->SSRC, 0, Timer::SESSMESS, 0,
	       TimerToRecycle->Sender, TimerToRecycle->SSRC, Laufzeit, false,
	       0, false, 0, TimerToRecycle);
    }
    break;
  }

  // *******************************************************************
  // RATECTL timer
  // *******************************************************************
  case Timer::RATECTL: {

#ifdef _SMP_RATECTRL
    fprintf(SMP::debug, "RATECTL-Timer ran out\n");
#endif

    RateControl(toHandle);
    SetTimer(-1, 0, 0, Timer::RATECTL, 0, NULL, 0, 0,
	     false, 0, false, 0, TimerToRecycle);
    break;
  }

  // *******************************************************************
  // REPEAT timer
  // *******************************************************************
  case Timer::REPEAT: {
    // 2. DelayedADU-Paket versenden
    // 3. Naechsten Repeat-Timer stellen

    // Mittlerweile kann ein Repeat gekommen sein und der Timer wurde nicht
    // ordentlich geloescht
    if (!c.OpenRequest[toHandle]) return;

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Client[%u]: Repeat Timer ran %u. time out\n",
	    toHandle, TimerToRecycle->i + 1);
#endif

    if (TimerToRecycle->i <= 4) { // nur fuenf mal wiederholen

#ifdef _SMP_LJ
      fprintf(SMP::debug, "\n");
#endif

      SendDADUReq(toHandle, c.DAduRequest[toHandle].data,
		  c.DAduRequest[toHandle].size, false);

      SetTimer(toHandle, c.We[toHandle]->SSRC,
	       0, Timer::REPEAT, DEFAULTDISTANCE,
	       c.We[toHandle], c.We[toHandle]->SSRC,
	       0, false, 0, 0, (TimerToRecycle->i + 1),
	       TimerToRecycle);
    } else {

#ifdef _SMP_LJ
      fprintf(SMP::debug, " - deleted\n");
#endif
      // neuer Request kann gestellt werden
      c.OpenRequest[toHandle] = false;

      // Anwendung benachrichtigen
      IndDADUReqFailed(toHandle);
    }
    break;
  }

  // *******************************************************************
  // GIVETHROUGH timer
  // *******************************************************************
  case Timer::GIVETHROUGH: {

    // *******************************************************************
    // find data location in DADU list
    // *******************************************************************
    DADU toFind(TimerToRecycle->SSRC, TimerToRecycle->SN);
    DADU* dadu = c.dadulist[toHandle].Find(&toFind);

    if (dadu) {

#ifdef _SMP_LJ
      fprintf(SMP::debug, "Demon::ExecuteTimerAction() : i = %u\n", TimerToRecycle->i);
#endif

     if (TimerToRecycle->i >= 1) {

       DADU * toDelete = c.dadulist[toHandle].Delete(dadu);
       if (toDelete)
	 delete(toDelete);
       else
	 fprintf(SMP::debug, "Demon::ExecuteTimerAction() : could not delete DADU request\n");
       delete(TimerToRecycle);

     } else {

       dadu->status = GT_TIMER_EXPIRED;
       IndDADUReq(toHandle, dadu->data, dadu->size, dadu->id, dadu->ip_addr);

#ifdef _SMP_LJ
       fprintf(SMP::debug, "Client [%u]: GiveTrough Timer ran out, application has been informed\n", toHandle);
#endif

       // set new GIVETHROUGH timer (will delete DADU)
       SetTimer(toHandle, TimerToRecycle->SSRC,
		TimerToRecycle->SN, Timer::GIVETHROUGH,
		0, 0, c.We[toHandle]->SSRC,
		0, true, 0, 0, (TimerToRecycle->i + 1), TimerToRecycle);
     }

   } else {

#ifdef _SMP_LJ
     fprintf(SMP::debug, "Demon::ExecuteTimerAction() : could not find DADU request for GIVETHROUGH operation (already discarded)\n");
#endif
     delete(TimerToRecycle);
   }


   break;
  }

  // *******************************************************************
  // DELAYEDSEND timer
  // *******************************************************************
  case Timer::DELAYEDSEND: {
    Operation toFind(TimerToRecycle->SN);

    Operation* Packet = TimerToRecycle->Sender->operationslist.Find(&toFind);
    if (Packet == NULL) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : could not find data for delayed send operation\n");
      return;
    }

    if (::send(c.msock_w[toHandle], Packet->data, Packet->size, 0) < 0) {
      fprintf(SMP::debug, "Demon::ExecuteTimerAction(%u) : send delayed failed\n");
    }

    fprintf(SMP::debug, "Demon::ExecuteTimerAction - send for client %u packet with SN %u\n", toHandle, TimerToRecycle->SN);

    delete TimerToRecycle;
    break;
  }


  // *******************************************************************
  // DELAYEDSEND UNRELIABLE timer
  // *******************************************************************
  case Timer::DELAYEDSENDU: {

    break;
  }


  // *******************************************************************
  // default
  // *******************************************************************
  default:
    fprintf(SMP::debug, "Demon::ExecuteTimerAction - Unknown ActionId\n");
    return;
  }
}

/*********************************************************************
 Name    : Demon::CheckLivenessOfMembers
 Param   :
 Result  :
 Desc    : checks activities of group members at time CHECKPERIOD
 Calledby: Demon::ExecuteTimerAction
 Uses    :
 *********************************************************************/
void SMP::Demon::CheckLivenessOfMembers(int toHandle) {

  // Funktion ueberprueft, ob die Mitglieder in der Liste
  // waehrend der letzten Check-Periode eine Sitzungsnachricht
  // gesendet haben. Es werden nur aktive Mitglieder, die ueber
  // die gesamte letzte Check-Periode bereits anwesend waren,
  // ueberprueft. Die Ueberpruefung findet nicht bei jedem Senden
  // einer SM an, sondern erst nach Ablauf einer Periode (CHECKPERIOD)

  int time = GetTimestamp();

  if ((++c.SMCount[toHandle] % CHECKPERIOD) == 0) {
    Member* ToCheck = c.grouplist[toHandle].First();
    while (ToCheck) {
      if (ToCheck->New)
	ToCheck->New = false;
      else {
	if (ToCheck->ActiveMember) {

	  if (ToCheck->LastSessionMessage
	      < c.TimeOfLastCheck[toHandle]) {
	    // Mitglied hat waehrend der letzten Periode keine
	    // Sitzungsnachricht mehr verschickt. Es muss von
	    // dem Tod des Mitglieds ausgegangen werden
	    ToCheck->ActiveMember = false;
	    if (!c.LateJoin[toHandle]) {
	      UpdateAllOpLists(toHandle, ToCheck->SSRC, 0, false);
	    }
	  }
	}
      }
      ToCheck = c.grouplist[toHandle].Next();
    }
    c.TimeOfLastCheck[toHandle] = time;
  }
  c.We[toHandle]->LastSessionMessage = time;
}


/*********************************************************************
 Name    : Demon::RateControl
 Param   :
 Result  :
 Desc    : controls rate of outgoing net traffic
 Calledby: Demon::ExecuteTimerAction
 Uses    :
 *********************************************************************/
void SMP::Demon::RateControl(int toHandle) {
  //
  // check credits for each client every second
  //
  for(int x = 0; x < MAX_CLIENTS; x++) {

#ifdef _SMP_RATECTRL
    int Kbps = (int)(c.SessionBandwidth[x] * 10 * 8 / 1000);
    int bytes_sent = (int)(c.SessionBandwidth[x] - c.Credit[x]);
    if (bytes_sent <= c.SessionBandwidth[x])
      Kbps = 8 * bytes_sent / 1000;

    if (Kbps > c.MaxKbps[x]) c.MaxKbps[x] = Kbps;

    fprintf(SMP::debug, "Demon::RateControl() : credit(%u) = %u bytes kbps = %u max = %u\n",
	    x, c.Credit[x], Kbps, c.MaxKbps[x]);
#endif

    c.Credit[x] = c.SessionBandwidth[x];

    BlockedADU* ADUToSend = c.ratecontrolqueue[x].First();
    while ((ADUToSend) && (c.Credit[x] > 0)) {

      switch(ADUToSend->Type) {

      case BlockedADU::ADU:
	if (SendData(x, ADUToSend->data, ADUToSend->size, false) < 0) {
	  fprintf(SMP::debug, "Demon::RateControl(%u) : could not send data\n", x);
	}
	break;
      case BlockedADU::URADU:
	if (SendUnreliableData(x, ADUToSend->data, ADUToSend->size, false) < 0) {
	  fprintf(SMP::debug, "Demon::RateControl(%u) : could not send data\n", x);
	}
	break;
      case BlockedADU::RADU:
	if (SendRepair(x, ADUToSend->data, ADUToSend->size, ADUToSend->Ttl) < 0) {
	  fprintf(SMP::debug, "Demon::RateControl(%u) : could not send repair\n", x);
	}
	break;
      default:
	fprintf(SMP::debug, "Demon::RateControl(%u) : unknown ADU type\n", x);
      }

      //
      // reduce credit
      //
      c.Credit[x] -= ADUToSend->size;

      delete(c.ratecontrolqueue[x].RemoveFirst());

      ADUToSend = c.ratecontrolqueue[x].First();
    }
  }
}


/*********************************************************************
 Name    : Demon::DeleteRunningTimer
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
void SMP::Demon::DeleteRunningTimer() {
  //struct itimerval LaufZeit;
  //timerclear(&LaufZeit.it_interval);
  //timerclear(&LaufZeit.it_value);
  //setitimer(ITIMER_REAL, &LaufZeit, NULL);
  Tcl_DeleteTimerHandler(timer);
}

/*********************************************************************
 Name    : Demon::FindTimer
 Param   :
 Result  :
 Desc    : gives back entry of timer list
 Calledby: Demon::receive
 Uses    :
 *********************************************************************/
SMP::Timer* SMP::Demon::FindTimer(int toHandle, int type,
			unsigned int ssrc, unsigned int SeqN) {

  Timer toSearch(0, ssrc, SeqN, type, 0, NULL,
                 c.We[toHandle]->SSRC,0, 0, 0, false, 0);
  Timer* Ret = timerlist.Find(&toSearch);
  return Ret;
}


/*********************************************************************
 Name    : Demon::DeleteAllTimerForLeavingApplication
 Param   :
 Result  :
 Desc    : Deletes all timers in list for leaving client / appl.
 Calledby: Demon::logOff
 Uses    :
 *********************************************************************/
void SMP::Demon::DeleteAllTimerForLeavingApplication(int toLogOff,
                                                unsigned int SSRC) {

  Timer toDie(0, 0, 0, -1,  0, NULL, SSRC, 0, 0, 0, false, 0); // x

#ifdef _SMP_SHOWLIST
  timerlist.Print();
#endif

  DeleteRunningTimer();

  // now delete all timers that match toDie
  timerlist.DeleteAll(&toDie);

  // take first remaining timer of list and reset itimer
  Current = timerlist.First();
  SignalHandler();

#ifdef _SMP_SHOWLIST
  timerlist.Print();
#endif

}


/*********************************************************************
 Name    : Demon::DeleteTimer
 Param   :
 Result  :
 Desc    : deletes timer in timer list
 Calledby: Demon::receive
 Uses    :
 *********************************************************************/
SMP::Timer* SMP::Demon::DeleteTimer(int toHandle, int type,
			  unsigned int ssrc, unsigned int SeqN) {


  // *******************************************************************
  // set delete mask
  // *******************************************************************
  Timer toDie(0, ssrc, SeqN, type, 0, NULL, c.We[toHandle]->SSRC,
	      0, 0, 0, false, 0);

  //
  // if toDie to delete equals the first timer in timer list, then we
  // have to call the signal handler in order to set the itimer to the
  // subsequent timer event
  //
  bool sig = false;
  Timer* Ret = timerlist.First();
  if (*Ret == toDie) {
    DeleteRunningTimer();
    Ret = timerlist.RemoveFirst();
    Current = timerlist.First();
    sig = true;
  } else Ret = timerlist.Delete(&toDie);


  // *******************************************************************
  // reset current timer
  // *******************************************************************
  if (sig) SignalHandler();

#ifdef _SMP_REQTIMER
  if (Ret) fprintf(SMP::debug, "deleting timer with id %u\n", Ret->ActionId);
#endif

  if (Ret && Ret->ActionId == Timer::SESSMESS)
    fprintf(SMP::debug, "deleting sessmess timer of client%u\n", toHandle);

  return Ret;
}


/*********************************************************************
 Name    : Demon::UpdateTimerList
 Param   :
 Result  :
 Desc    : changes SSRC of client / application in the case of a
           SSRCCollision
 Calledby: Demon::SSRCCollision
 Uses    :
 *********************************************************************/
void SMP::Demon::UpdateTimerList(unsigned int OldSSRC, unsigned int NewSSRC) {

  Timer* t = timerlist.First();
  while (t != NULL) {
        if (t->Erzeuger == OldSSRC)
            t->Erzeuger = NewSSRC;
         t = timerlist.Next();
  }
}

/*********************************************************************
 Name:      Demon::OpenReceiveSocket(LogOnInfo info)
 Param:     LogOnInfo
 Result:    -1 om failure, otherwise 1
 Desc:      opens receive socket for client toHandle
 Calledby:  logOn
 *********************************************************************/
int SMP::Demon::OpenReceiveSocket(int toHandle, LogOnInfo* info) {


  if ((c.msock_r[toHandle] = rx_sock_init(info->multicastAddr,
					  info->multicastPort,
					  info->ttl)) < 0) {
    fprintf(SMP::debug, "Demon::OpenReceiveSocket - function socket cannot access multicast socket\n");
#ifdef WIN32
    fprintf(SMP::debug, "%u\n", WSAGetLastError());
#endif
    fprintf(SMP::debug, "socket is being closed - try to connect again\n");
    logOff(toHandle);
    Demon::close(toHandle);
    return -1;
  }

  int bufsize = 80000;
  if (setsockopt(c.msock_r[toHandle], SOL_SOCKET, SO_RCVBUF,
		 (const char*)&bufsize, sizeof(bufsize)) < 0) {
#ifdef WIN32
    fprintf(SMP::debug, "%u\n", WSAGetLastError());
#endif
    fprintf(SMP::debug, "Demon::OpenReceiveSocket - function setsockopt (SO_RCVBUF) cannot access multicast socket\n");
    return -1;
  }
  return 1;
}

/*********************************************************************
 Name:      Demon::OpenSendSocket()
 Param:     LogOnInfo
 Result:    -1 on failure otherwise 1
 Desc:      opens send socket, sets loop back option
 Calledby:  logOn
 *********************************************************************/
int SMP::Demon::OpenSendSocket(int toHandle, LogOnInfo* info) {

  if((c.msock_w[toHandle] = tx_sock_init(info->multicastAddr,
					 info->multicastPort,
					 info->ttl)) < 0) {
    fprintf(SMP::debug, "Demon::OpenSendSocket - function socket cannot access multicast socket\n");
#ifdef WIN32
    fprintf(SMP::debug, "%u\n", WSAGetLastError());
#endif
    fprintf(SMP::debug, "socket is being closed - try to connect again\n");
    logOff(toHandle);
    Demon::close(toHandle);
    return -1;
  }

  int bufsize = 80000;
  if (setsockopt(c.msock_w[toHandle], SOL_SOCKET, SO_SNDBUF,
		 (const char*)&bufsize, sizeof(bufsize)) < 0) {
    fprintf(SMP::debug, "Demon::OpenReceiveSocket - function setsockopt (SO_SNDBUF) cannot access multicast socket\n");
#ifdef WIN32
    fprintf(SMP::debug, "%u\n", WSAGetLastError());
#endif
    return -1;
  }


#ifndef WIN32
  if (c.Multicast[toHandle]) {
    //if (!mc_loop) {

      char loop = 1;
      if (setsockopt(c.msock_w[toHandle], IPPROTO_IP,
		     IP_MULTICAST_LOOP, (const char*)&loop, sizeof(loop)) < 0) {
	fprintf(SMP::debug, "Demon::OpenSendSocket - function setsockopt (IP_MULTICAST_LOOP) cannot access multicast socket\n");
	fprintf(SMP::debug, "socket is being closed - try to connect again\n");;
	logOff(toHandle);
	Demon::close(toHandle);
	return -1;
      }

      //}
  }
#endif

  return 1;
}

/*********************************************************************
 Name:      Demon::Reconnect()
 Param:     client toHandle
 Result:    -1, 1
 Desc:      uncast connections elicit an ECONNREFUSED error if no
            recipant of the data exists, thus, try to reconnect until
            a partner becomes available, but don't exit
 *********************************************************************/
int SMP::Demon::Reconnect(int toHandle) {

  //struct sockaddr_in address;
  //bzero(&address, sizeof(address));

  /* we do not need to reconnect by now, keep this a while */
  /* disconnect */
  //address.sin_family = AF_UNSPEC;
  //if (connect(c.msock_w[toHandle], (struct sockaddr *)&address,
  //	      sizeof(addr)) == -1) {
  //  if (errno != EAFNOSUPPORT) {
  //    perror("Demon::Reconnect() ; could not disconnect socket");
  //    return -1;
  //  }
  //}

  /* connect */
  //c.mc_addr_w[toHandle].sin_family = AF_INET;
  //if (connect(c.msock_w[toHandle], (struct sockaddr *) &c.mc_addr_w[toHandle],
  //	      sizeof(addr)) == -1) {
  //  perror("Demon::OpenSendSocket - function connect cannot access "
  //	   "multicast socket");
  //  cerr << "socket is being closed - try to connect again" << endl;
  //  return -1;
  //} else {
  //  return 1;
  //  }*/
  return 1;
}

/*********************************************************************
 Name:      Demon::sendJoinMsg()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendJoinMsg(int toHandle) {
  SMPHeader Header;
  u_int16 flags = 0 << 15 | SMPHeader::JOIN;
  Header.FLAGS = htons(flags);
  Header.SSRC = htonl(c.We[toHandle]->SSRC);
  Header.ORGIP = HostIpAddr;
  Header.TS   = htonl(GetTimestamp());
  Header.SZ   = htons(sizeof(SMPHeader));
  if (sendCtlMsg(toHandle, &Header, sizeof(SMPHeader)))
    return(sizeof(SMPHeader));
  else
    return(-1);
}

/*********************************************************************
 Name:      Demon::SendLeaveMsg()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendLeaveMsg(int toHandle) {
  SMPHeader Header;
  u_int16 flags = 0 << 15 | SMPHeader::LEAV;
  Header.FLAGS = htons(flags);
  Header.SSRC = htonl(c.We[toHandle]->SSRC);
  Header.ORGIP = HostIpAddr;
  Header.TS   = htonl(GetTimestamp());
  Header.SZ   = htons(sizeof(SMPHeader));
  if (sendCtlMsg(toHandle, &Header, sizeof(SMPHeader)))
    return(sizeof(SMPHeader));
  else
    return(-1);
}

/*********************************************************************
 Name:      Demon::SendData()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendData(int toHandle, void* data, int size, bool block) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendData(%u) : client does not exist\n", toHandle);
    return(-1);
  }

#ifdef _SMP_DEBUG
  static unsigned int count = 1;
  fprintf(SMP::debug, "Demon::SendData - send for Client %u bytes %u nbr %u\n", toHandle, size, count);
  ++count;
#endif

  // *******************************************************************
  // check credit
  // *******************************************************************
  if (c.Credit[toHandle] <= 0) {
    BlockedADU* packet = new BlockedADU(BlockedADU::ADU, data, size);

    if (!packet) {
      fprintf(SMP::debug, "Demon::SendData(%u) : operation new BlockedADU failed socket is being closed - try to connect again\n", toHandle);
      logOff(toHandle);
      Demon::close(toHandle);
      return(-1);
    }

#ifdef _SMP_RATECTRL
    fprintf(SMP::debug, "Demon::sendData(%u) : insert packet in rate control list\n", toHandle);
#endif

    c.ratecontrolqueue[toHandle].Insert(packet);
    return(0);
  }

  // *******************************************************************
  // build header
  // *******************************************************************
  SMPHeader* Header = (SMPHeader *)data;
  u_int16 flags = 0 << 15 | SMPHeader::DATA; /* X = 0, PT = 0 */
  Header->FLAGS = htons(flags);
  Header->SSRC = htonl(c.We[toHandle]->SSRC);
  Header->ORGIP = HostIpAddr;       /* defaults to network byte order ok */
  Header->TS = htonl(GetTimestamp());
  Header->SZ = htons(size);
  Header->SN = htonl(GetSequenceNumber(toHandle));

  // *******************************************************************
  // send data
  // *******************************************************************

#ifdef _SMP_SENDDELAYED
  c.We[toHandle]->SN++;
  if (!(c.MSet[toHandle]->Find(c.We[toHandle]))) c.MSet[toHandle]->Insert(c.We[toHandle]);
  if (!InsertOpList(toHandle, c.We[toHandle], data, size, c.We[toHandle]->SN, block))
    fprintf(SMP::debug, "Demon::SendData() : could not insert packet in op list\n");
  SetTimer(toHandle, c.We[toHandle]->SSRC, c.We[toHandle]->SN, Timer::DELAYEDSEND, 0,
	   c.We[toHandle], Deamon->c.We[toHandle]->SSRC, 0, false, 0, false, 0);
  c.Credit[toHandle] -= size;
  fprintf(SMP::debug, "Demon::SendData - buffer for client %u packet with SN %u\n", toHandle, c.We[toHandle]->SN);
  return size;
#endif

#ifdef _SMP_ERROR
  static std::bitset<100000> erstes[5];
  if ((!erstes[toHandle].test(ntohl(Header->SN))) && (ntohl(Header->SN) % 3) == 0) {
    erstes[toHandle].set(ntohl(Header->SN));
  } else {
#endif

#ifdef _SMP_SORDER
    fprintf(SMP::debug, "Demon::sendData(%u) : send packet no. %u\n", toHandle, ntohl(Header->SN));
#endif

    if (::send(c.msock_w[toHandle], (char*)data, size, 0) < 0) {

#ifdef WIN32
      if ((WSAGetLastError() == WSAECONNREFUSED) && !c.Multicast[toHandle]) {
	Reconnect(toHandle);
      } else {
	fprintf(SMP::debug, "Demon::SendData - function send cannot access mutlicast socket%u \n ", WSAGetLastError());
	logOff(toHandle);
	Demon::close(toHandle);
	return(-1);
      }
#else
      //UNIX
      /* we are trying to send unicast data with no recipients => reconnect */
      if ((errno == ECONNREFUSED) && (!c.Multicast[toHandle])) {
	Reconnect(toHandle);
      } else {
	fprintf(SMP::debug, "Demon::send - function send cannot access multicast socket\n");
	fprintf(SMP::debug, "socket is being closed - try to connect again %u\n", errno);
	logOff(toHandle);
	Demon::close(toHandle);
	return(-1);
      }
#endif
    }

#ifdef _SMP_ERROR
  }
#endif

#ifdef _SMP_LJ
  if (toHandle == 2) {
    fprintf(SMP::debug, "Demon::SendData(%u)\n", ntohl(Header->SN));
  }
#endif


  // *******************************************************************
  // update credit & sequence number
  // *******************************************************************
  c.Credit[toHandle] -= size;
  c.We[toHandle]->SN++;

  // *******************************************************************
  // insert adu in operations list
  // *******************************************************************
  if (!(c.MSet[toHandle]->Find(c.We[toHandle])))
    c.MSet[toHandle]->Insert(c.We[toHandle]);

  if (!InsertOpList(toHandle, c.We[toHandle], data, size, c.We[toHandle]->SN, block))
    fprintf(SMP::debug, "Demon::SendData() : could not insert packet in op list\n");

  return(size);
}


/*********************************************************************
 Name:      Demon::SendUnreliableData()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendUnreliableData(int toHandle, void* data, int size, bool block) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendData(%u) : client does not exist\n", toHandle);
    return(-1);
  }

  if (toHandle == 2)
    fprintf(SMP::debug, "Demon::SendUnreliableData - send for Client %u bytes %u\n", toHandle, size);

  // *******************************************************************
  // check credit
  // *******************************************************************
  if (c.Credit[toHandle] <= 0) {
    BlockedADU* packet = new BlockedADU(BlockedADU::URADU, data, size);

    if (!packet) {
      fprintf(SMP::debug, "Demon::SendData(%u) : operation new BlockedADU failed\n", toHandle);
      fprintf(SMP::debug, "socket is being closed - try to connect again\n");
      logOff(toHandle);
      Demon::close(toHandle);
      return(-1);
    }

#ifdef _SMP_RATECTRL
    fprintf(SMP::debug, "Demon::sendUnreliableData(%u) : insert packet in rate control list\n", toHandle);
#endif

    c.ratecontrolqueue[toHandle].Insert(packet);
    return(0);
  }

  // *******************************************************************
  // build header
  // *******************************************************************
  SMPHeader* Header = (SMPHeader *)data;
  u_int16 flags = 0 << 15 | SMPHeader::DATA_UNRELIABLE; /* X = 0, PT = 1 */
  Header->FLAGS = htons(flags);
  Header->SSRC = htonl(c.We[toHandle]->SSRC);
  Header->ORGIP = HostIpAddr;       /* defaults to network byte order ok */
  Header->TS = htonl(GetTimestamp());
  Header->SZ = htons(size);
  Header->SN = 0; // no sequence numbers

  // *******************************************************************
  // send data
  // *******************************************************************
  if (::send(c.msock_w[toHandle], (char*)data, size, 0) < 0) {
#ifdef WIN32
    if ((WSAGetLastError() == WSAECONNREFUSED) && !c.Multicast[toHandle]) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::SendUnreliableData - function send cannot access socket %u \n", WSAGetLastError());
      logOff(toHandle);
      Demon::close(toHandle);
      return -1;
    }
#else
    // UNIX
    /* we are trying to send unicast data with no recipients => reconnect */
    if ((errno == ECONNREFUSED) && (!c.Multicast[toHandle])) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::SendUnreliableData - function send cannot access multicast socket\n");
      fprintf(SMP::debug, "socket is being closed - try to connect again %u\n", errno);
      logOff(toHandle);
      Demon::close(toHandle);
      return(-1);
    }
#endif
  }

#ifdef _SMP_UNRELIABLE
  fprintf(SMP::debug, "Demon::SendUnreliableData - sent %u bytes\n", size);
#endif

  // *******************************************************************
  // update credit
  // *******************************************************************
  c.Credit[toHandle] -= size;

  return(size);
}


/*********************************************************************
 Name:      Demon::SendDADUReq()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendDADUReq(int toHandle, void* data, int size, bool block) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendDADUReq(%u) : client does not exist\n", toHandle);
    return(-1);
  }

  // *******************************************************************
  // build header
  // *******************************************************************
  SMPHeader* Header = (SMPHeader *)data;
  u_int16 flags = 0 << 15 | SMPHeader::DADURQ; /* X=0, PT=100 */
  Header->FLAGS = htons(flags);
  Header->SSRC  = htonl(c.We[toHandle]->SSRC);
  Header->ORGIP = HostIpAddr;
  Header->TS    = htonl(GetTimestamp());
  Header->SZ    = htons(size);
  Header->SN    = htonl(c.DAduRequest[toHandle].sn);

#ifdef _SMP_LJ
  fprintf(SMP::debug, "Client[%u]: Sending DelayedADURequest %u bytes at position %u\n", size, (unsigned long)(data));
#endif

#ifdef _SMP_SHOWLIST
  timerlist.Print();
#endif

  if (::send(c.msock_w[toHandle], (char*)data, size, 0) < 0) {
#ifdef WIN32
    if ((WSAGetLastError() == WSAECONNREFUSED) && !c.Multicast[toHandle]) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::SendDelayedADURequest - function send cannot access socket %u\n", WSAGetLastError());
      logOff(toHandle);
      Demon::close(toHandle);
      return -1;
    }
#else
    // UNIX
    /* we are trying to send unicast data with no recipients => reconnect */
    if ((errno == ECONNREFUSED) && (!c.Multicast[toHandle])) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::SendDelayedADURequest - function send cannot access multicast socket\n");
      fprintf(SMP::debug, "socket is being closed - try to connect again %u \n", errno);
      logOff(toHandle);
      Demon::close(toHandle);
      return(-1);
    }
#endif
  }

  //
  //  if (block)

  //    c.Credit[toHandle] -= size;

  return 1;
}

/*********************************************************************
 Name:      Demon::SendDADURsp()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendDADURsp(int toHandle, void* data, int size, unsigned int id,
		       bool block) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendDADURsp(%u) : client does not exist\n", toHandle);
    return(-1);
  }

  DADU * dadu = c.FindDADU(toHandle, id);

  if (dadu) {

    // *******************************************************************
    // build header
    // *******************************************************************
    SMPHeader* Header = (SMPHeader *) data;
    u_int16 flags = 0 << 15 | SMPHeader::DADURP; /* X=0, PT = 101 */
    Header->FLAGS = htons(flags);
    Header->SSRC = htonl(c.We[toHandle]->SSRC);
    Header->ORGIP = HostIpAddr;
    Header->TS   = htonl(GetTimestamp());
    Header->SZ   = htons(size);
    Header->SN = htonl(dadu->sn);

    // *******************************************************************
    // set RSSRC
    // *******************************************************************
    SMPDelADURespBody* delADUResp =
      (SMPDelADURespBody*)(((char*)data) + sizeof(SMPHeader));

    delADUResp->RSSRC = htonl(dadu->ssrc);

#ifdef _SMP_LJ
   fprintf(SMP::debug, "Client[%u] >>>>Send DelayedADU-Response for client/sn %u / %u, %u Bytes at Position %u\n",
	   toHandle, dadu->ssrc, dadu->sn, size, (unsigned long)(data));
#endif

#ifdef _SMP_SHOWLIST
    timerlist.Print();
#endif

    if (::send(c.msock_w[toHandle],(char*)data, size, 0) < 0) {
#ifdef WIN32
      if ((WSAGetLastError() == WSAECONNREFUSED) && !c.Multicast[toHandle]) {
	Reconnect(toHandle);
      } else {
	fprintf(SMP::debug, "Demon::SendDelayedADUResponse - function send cannot access socket %u\n", WSAGetLastError());
	logOff(toHandle);
	Demon::close(toHandle);
	return -1;
      }
#else
      // UNIX
      /* we are trying to send unicast data with no recipients => reconnect */
      if ((errno == ECONNREFUSED) && (!c.Multicast[toHandle])) {
	Reconnect(toHandle);
      } else {
	fprintf(SMP::debug, "Demon::SendDelayedADUResponse - function send cannot access multicast socket\n");
	fprintf(SMP::debug, "socket is being closed - try to connect again %u\n", errno);
	logOff(toHandle);
	Demon::close(toHandle);
	return(-1);
      }
#endif
    }

    FreePacketBuffer(toHandle, (char*)data, size);

    dadu->reply_time = GetTimestamp();
    // *******************************************************************
    // delete DADU request; not needed anymore since req is answered now
    // *******************************************************************
    // DADU * toDelete = c.dadulist[toHandle].Delete(dadu);
    //if (toDelete)
    // delete(toDelete);
    //  else
    //  cerr << "Demon::SendDADURsp() : could not delete DADU request" << endl;

    if (block) c.Credit[toHandle] -= size;
    return(size);

  } else {

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Demon::SendDADURsp() : could not find DADU request\n");
#endif

    return(-1);
  }
}

/*********************************************************************
 Name:      Demon::SendRepair()
 Param:
 Result:
 Desc:
 Calledby:  signalHandler()
 Uses:
 *********************************************************************/
int SMP::Demon::SendRepair(int toHandle, void * data, int size, unsigned char TTL) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendRepair(%u) : client does not exist", toHandle);
    return(-1);
  }

  // *******************************************************************
  // change header
  // *******************************************************************
  SMPHeader* Header = (SMPHeader *)data;
  u_int16 flags = 1 << 15 | 0;  /* X = 1,  PT = 0 */
  Header->FLAGS = htons(flags);
  Header->SZ = htons(size);

  // *******************************************************************
  // don't send data if credit is below zero
  // instead put data into a queue, which is processed
  // by the RateCtrl function
  // *******************************************************************
  if (c.Credit[toHandle] <= 0) {
    BlockedADU* packet = new BlockedADU(BlockedADU::RADU, data, size, TTL);
    if (!packet) {
      fprintf(SMP::debug, "Demon::SendRepair - operation new BlockedADU failed\n");
      fprintf(SMP::debug, "socket is being closed - try to connect again\n");
      logOff(toHandle);
      Demon::close(toHandle);
      return (-1);
    }

#ifdef _SMP_RATECTRL
    fprintf(SMP::debug, "Demon::sendRepair(%u) : insert packet in rate control list\n", toHandle);
#endif

    c.ratecontrolqueue[toHandle].AddFirst(packet);
    return(0);
  }

#ifdef _SMP_SORDER
  fprintf(SMP::debug, "Demon::sendRepair(%u) : send repair packet no. %u\n", toHandle, ntohl(Header->SN));
#endif

  // *******************************************************************
  // send REPAIR
  // *******************************************************************
  if (::send(c.msock_w[toHandle], (char*)data, size, 0) < 0) {
#ifdef WIN32
    if ((WSAGetLastError() == WSAECONNREFUSED) && !c.Multicast[toHandle]) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::SendRepair - function send cannot access socket %u\n", WSAGetLastError());
      logOff(toHandle);
      Demon::close(toHandle);
      return -1;
    }
#else
    // UNIX
    /* we are trying to send unicast data with no recipients => reconnect */
    if ((errno == ECONNREFUSED) && (!c.Multicast[toHandle])) {
      Reconnect(toHandle);
    } else {
      fprintf(SMP::debug, "Demon::SendRepair - function send cannot access multicast socket\n");
      fprintf(SMP::debug, "socket is being closed - try to connect again %u\n", errno);
      logOff(toHandle);
      Demon::close(toHandle);
      return(-1);
    }
#endif
  }
  return(size);
}

/*********************************************************************
 Name:      Demon::SendRequest()
 Param:
 Result:
 Desc:
 Calledby:  signalHandler()
 Uses:
 *********************************************************************/
int SMP::Demon::SendRequest(int toHandle, unsigned int req_ssrc,
		       unsigned int req_sn) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendRequest(%u) : client does not exist\n", toHandle);
    return(-1);
  }

  char RequestBuffer[sizeof(SMPHeader) + sizeof(SMPRequestBody)];

  SMPHeader* Header = (SMPHeader *)RequestBuffer;
  u_int16 flags = 0 << 15 | SMPHeader::REQ;
  Header->FLAGS = htons(flags);
  Header->SSRC = htonl(Deamon->c.We[toHandle]->SSRC);
  Header->ORGIP = HostIpAddr;
  Header->TS   = htonl(Deamon->GetTimestamp());
  Header->SZ   = htons(sizeof(SMPHeader) + sizeof(SMPRequestBody));
  SMPRequestBody* RequestBody = (SMPRequestBody *)
    (RequestBuffer + sizeof(SMPHeader));
  RequestBody->ReqSSRC = htonl(req_ssrc);
  RequestBody->ReqSN = htonl(req_sn);
  RequestBody->ReqTTL = htons(Deamon->c.GlobalTTL[toHandle]);

  int size = sizeof(SMPHeader) + sizeof(SMPRequestBody);
#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Demon::SendRequest - size %u ReqSSRC %u\n", toHandle, ntohl(RequestBody->ReqSSRC));
#endif
  Deamon->sendCtlMsg(toHandle, RequestBuffer, size, false);
  return(size);
}

/*********************************************************************
 Name:      Demon::SendSessionMsg()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::_SendSessionMsg(int toHandle, unsigned int we_ssrc,
			  unsigned int we_sn) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendSessionMsg(%u) : client does not exist\n", toHandle);
    return(-1);
  }

  //char MessageBuffer[sizeof(SMPHeader) + sizeof(int) +
	//	    MAX_MEMBERS_PER_SESSION_MESSAGE *
	//	    sizeof(SMPSessionMessageElement)];

  // *******************************************************************
  // build header
  // *******************************************************************
  SMPHeader* Header = (SMPHeader *)send_buf;
  u_int16 flags = 0 << 15 | SMPHeader::SMES;
  Header->FLAGS = htons(flags);
  Header->SSRC = htonl(we_ssrc);
  Header->ORGIP = HostIpAddr;
  Header->SN   = htonl(we_sn);
  Header->TS   = htonl(GetTimestamp());

  // *******************************************************************
  // build body
  // *******************************************************************
  int BodyLen = AddSessionMessageBody(send_buf, toHandle);

#ifdef _SMP_SESS
  fprintf(SMP::debug, "Demon::SendSessionMsg(%u) : send sessmess with seq.no. = %u\n", toHandle, we_sn);
#endif

  Header->SZ = htons(sizeof(SMPHeader) + BodyLen);

#ifdef _SMP_HEADER
  fprintf(SMP::debug, "Demon::_SendSessionMessage - sending... X = %u PT = %u SN = %u SZ = %u SSRC = %u\n",
	  ntohs(Header->FLAGS)&SMP_X, ntohs(Header->FLAGS)&SMP_PT, ntohl(Header->SN), ntohs(Header->SZ), ntohl(Header->SSRC));
#endif

  sendCtlMsg(toHandle, send_buf, BodyLen + sizeof(SMPHeader), false);
  return(BodyLen + sizeof(SMPHeader));
}


/*********************************************************************
 Name:      Demon::SendSessionMsg()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::SendSessionMsg(int toHandle, unsigned int we_ssrc,
			  unsigned int we_sn) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::SendSessionMsg(%u) : client does not exist\n");
    return(-1);
  }

  /* cycle through all members */
  Member* MemberToSend = c.MSet[toHandle]->First();
  int MAXMEMBERS = c.MSet[toHandle]->GetMaxCount();
  SMPSessionMessageElement * Element;
  SMPSessionMessageCount* Count;
  SMPHeader* Header;
  bool new_msg = true;
  int k = 0;
  int j = 0;


  do {

    if(new_msg) {
      // *******************************************************************
      // build header
      // *******************************************************************
      Header = (SMPHeader *)send_buf;
      u_int16 flags = 0 << 15 | SMPHeader::SMES;
      Header->FLAGS = htons(flags);
      Header->SSRC = htonl(we_ssrc);
      Header->ORGIP = HostIpAddr;
      Header->SN   = htonl(we_sn);
      Header->TS   = htonl(GetTimestamp());
      Element = (SMPSessionMessageElement *)(send_buf + (sizeof(SMPHeader) + sizeof(SMPSessionMessageCount)));
      new_msg = false;
    }

    if (MemberToSend) {
      Element->SSRC = htonl(MemberToSend->SSRC);
      Element->SN = htonl(MemberToSend->SN);
      Element->HSN = htonl(MemberToSend->HighestNum);
      Element->Timestamp1 = htonl(MemberToSend->timestampSent);
      Element->Delta = htonl((GetTimestamp()
			      - MemberToSend->timestampReceived));
      Element->Active = htons(MemberToSend->ActiveMember);

      k += 1;
      Element += 1;
      MemberToSend = c.MSet[toHandle]->Next();
    }

    //    cout << "k= " << k << " MAXMEMBERS = " << MAXMEMBERS << " MemberToSend = "
    //	 << (unsigned int)MemberToSend << endl;


    if ((k == MAXMEMBERS) || (MemberToSend == NULL)) {
      Count = (SMPSessionMessageCount *)(send_buf + sizeof(SMPHeader));
      Count->i = htonl(k);
      int BodyLen =  sizeof(SMPSessionMessageCount)
	+ k*sizeof(SMPSessionMessageElement);
      Header->SZ = htons(sizeof(SMPHeader) + BodyLen);

#ifdef _SMP_HEADER
  fprintf(SMP::debug, "Demon::_SendSessionMessage - sending... X = %u PT = %u SN = %u SZ = %u SSRC = %u\n",
	  ntohs(Header->FLAGS)&SMP_X, ntohs(Header->FLAGS)&SMP_PT, ntohl(Header->SN), ntohs(Header->SZ), ntohl(Header->SSRC));
#endif

#ifdef _SMP_SESS
  fprintf(SMP::debug, "Demon::SendSessionMsg(%u) : send sessmess (size %u) with seq no %u number of members %u\n",
	  BodyLen + sizeof(SMPHeader), we_sn, ntohl(Count->i));
#endif

      sendCtlMsg(toHandle, send_buf, BodyLen + sizeof(SMPHeader), false);
      j++;
      if (MemberToSend) {
	k = 0;
	new_msg = true;
      }
    }
  } while(MemberToSend);

  //cout << "j = " << j << endl;

  return j;
}

/*********************************************************************
 Name    : Demon::AddSessionMessageBody
 Param   :
 Result  :
 Desc    :
 Calledby: Demon::Send SessionMsg()
 Uses    :
 *********************************************************************/
int SMP::Demon::AddSessionMessageBody(char* Header, int toHandle) {

  SMPSessionMessageElement* Element = (SMPSessionMessageElement *)
    (Header + (sizeof(SMPHeader) + sizeof(SMPSessionMessageCount)));

  Member* MemberToSend = c.MSet[toHandle]->First();
  int MAXMEMBERS = c.MSet[toHandle]->GetMaxCount();

  int k = 0;
  while ((MemberToSend != NULL) && (k < MAXMEMBERS)) {

    Element->SSRC = htonl(MemberToSend->SSRC);
    Element->SN = htonl(MemberToSend->SN);
    Element->HSN = htonl(MemberToSend->HighestNum);
    Element->Timestamp1 = htonl(MemberToSend->timestampSent);
    Element->Delta = htonl((GetTimestamp()
			    - MemberToSend->timestampReceived));
    Element->Active = htons(MemberToSend->ActiveMember);
    k += 1;
    Element += 1;
    MemberToSend = c.MSet[toHandle]->Next();
  }

  SMPSessionMessageCount* Count = (SMPSessionMessageCount *)
    (Header + sizeof(SMPHeader));

  Count->i = htonl(k);
  return (sizeof(SMPSessionMessageCount) + k*sizeof(SMPSessionMessageElement));
}

/*********************************************************************
 Name:      Demon::IndDADUReq()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
void SMP::Demon::IndDADUReq(int toHandle, char* data, int size, unsigned int id, unsigned int senderIPAddr) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::DADUInd(%u) : client does not exist\n");
    return;
  }

#ifdef _SMP_DATATRACE
  fprintf(SMP::debug, "Demon::DADUReqInd()\n");
#endif

  int _size = size - sizeof(SMPHeader);

  Indication ind;
  ind.buf  = new Buffer(_size);
  ind.ServiceType = Indication::DADUIND;
  ind.senderIPAddr = senderIPAddr;
  ind.Id = id;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  memcpy(ind.buf->getCont(), data + sizeof(SMPHeader), _size);

  SMP::smp_pointer->group[toHandle].Callback(ind);
}


/*********************************************************************
 Name:      Demon::IndDADURsp()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
void SMP::Demon::IndDADURsp(int toHandle, char* data, int size, unsigned int senderIPAddr) {

  if (!c.We[toHandle]) {
    fprintf(SMP::debug, "Demon::DADURsp(%u) : client does not exist\n");
    return;
  }

#ifdef _SMP_DATATRACE
  fprintf(SMP::debug, "Demon::DADURspInd()\n");
#endif

  int _size = size - sizeof(SMPHeader) - sizeof(SMPDelADURespBody);

  Indication ind;
  ind.buf  = new Buffer(_size);
  ind.ServiceType = Indication::DADUCONF;
  ind.senderIPAddr = senderIPAddr;
  ind.param = SMP::smp_pointer->group[toHandle].Param;

  memcpy(ind.buf->getCont(), data + sizeof(SMPHeader)+ sizeof(SMPDelADURespBody), _size);

  SMP::smp_pointer->group[toHandle].Callback(ind);
}


/*********************************************************************
 Name:
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
SMP::Member* SMP::Demon::CheckMember(int toHandle, unsigned int ssrc, unsigned int ip,
			   int port, unsigned int ts, int pt, bool x,
			   bool* unknown) {

  Member ToFind(ssrc);
  Member* MessageSender = c.grouplist[toHandle].Find(&ToFind);

  // *******************************************************************
  // could not find member
  // *******************************************************************
  if (MessageSender == NULL) {

    *unknown = true;

    if (c.CollisionList[toHandle].find(ssrc) != c.CollisionList[toHandle].end()) {
      // Paket von einer kollidierten Datenquelle
      // Paket Ignorieren
      return(NULL);

    } else {

      // *******************************************************************
      // create new member
      // *******************************************************************
      if ((pt != SMPHeader::LEAV) && (pt != SMPHeader::REQ) &&
	  (pt != SMPHeader::DADURQ) && (pt != SMPHeader::DADURP)) {

	MessageSender = new Member(ssrc, ip, port);
	if (!MessageSender) {
	  fprintf(SMP::debug, "Demon::CheckMember() - operation new Member failed\n");
	  return(NULL);
	}

	MessageSender->timestampSent = ts;
	unsigned int time = GetTimestamp();
	MessageSender->timestampReceived = time;
	MessageSender->LastSessionMessage = time;
	MessageSender->delay = DEFAULTDISTANCE;
	MessageSender->ActiveMember = true;

	if (!(c.MSet[toHandle]->Find(c.We[toHandle])))
	  c.MSet[toHandle]->Insert(MessageSender);

	c.Members[toHandle]++;
	c.grouplist[toHandle].Insert(MessageSender);

#ifdef _SMP_DEBUG
	c.grouplist[toHandle].Print();
#endif

      } else {
	return(NULL);
      }
    }
  } else {

    // SSRC-Kollisionen erkennen und behandeln
    if (MessageSender->IpAddr == 0) {
      if (!x)
	MessageSender->IpAddr = ip;
    } else {
      if (MessageSender->Port == 0) {
	if (!x) {
	  MessageSender->Port = port;
	}
      } else {

	// Vergleich kann statt finden
	if ((!x) && ((MessageSender->IpAddr != ip) ||
		     (MessageSender->Port != port))) {
	  // Kollision
	  #ifdef _SMP_DEBUG
	  fprintf(SMP::debug, ">>>>> SSRC-Collision detected\n");
	  #endif
	  SSRCCollision(toHandle, ssrc);
	  return(NULL);
	}
      }
    }
  }
  return(MessageSender);
}

/*********************************************************************
 Name:      Demon::ReceiveData()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveData(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;
  unsigned int OldSN;

  // *******************************************************************
  // read header data
  // *******************************************************************
  SMPHeader* Header = (SMPHeader*)data;

  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;
  unsigned int orgIP = Header->ORGIP;

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {
    //if (c.Loopback[toHandle]) { // client wishes to receive his own packets
    //IndData(toHandle, data, bread, orgIP);
    FreePacketBuffer(toHandle, data, bread);
    return(0);
    //} else {
    //FreePacketBuffer(toHandle, data, bread);
    //return(0);
    //}
  }

  // fprintf(SMP::debug, "Demon::ReceiveData - for client %u %u bytes\n", toHandle, bread);

  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

  if (!receivedX)
    Header->FLAGS = htons(0 << 15 | ntohs(Header->FLAGS)&SMP_PT);

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Demon::ReceiveData - Client[%u]: receive from Member %u packet with SN %u size %u\n",
	  toHandle, receivedSSRC, receivedSN, bread);
#endif

  // *******************************************************************
  // check member (check must not be interrupted!)
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);
  if (!MessageSender) {
    FreePacketBuffer(toHandle, data, bread);
    fprintf(SMP::debug, "Demon::receive(%u) : member error\n");
    return(-1);
  }
  if (senderUnKnown) {
    // *******************************************************************
    // update sequence number of message sender
    // *******************************************************************
    OldSN = MessageSender->SN;
    MessageSender->SN = receivedSN;
  }

#ifdef _SMP_LJ
  if (toHandle == 2) {
    fprintf(SMP::debug, "Demon::receive(%u) : received lj data\n", receivedSN);
  }
#endif

  // *******************************************************************
  // start packet processing
  // *******************************************************************
#ifdef _SMP_SORDER
  fprintf(SMP::debug, "Client %u RSN/HN/SN = %u/%u/%u\n",
	  toHandle, receivedSN, MessageSender->HighestNum, MessageSender->SN);
#endif

  // *******************************************************************
  // unknown sender - first packet of new sender
  // *******************************************************************
  if (senderUnKnown) {

#ifdef _SMP_SORDER
    fprintf(SMP::debug, "\t   case 1: Sender unkown\n");
#endif

    // *******************************************************************
    // packets are missing: req timer if smp late join or LJIND
    // *******************************************************************
    if (receivedSN > 1) {

      if (!c.LateJoin[toHandle]) {

	// Anwendung benachrichtigen
	if (c.FirstLateJoinInd[toHandle]) {

#ifdef _SMP_LJ
	  fprintf(SMP::debug, "Demon::ReceiveData() : received data, send LJIND to app toHandle = %u\n", toHandle);
#endif

	  IndLateJoin(toHandle, receivedSSRC);
	  c.FirstLateJoinInd[toHandle] = false;
	}

	// regulaeren Repair-Mechanismus unterdruecken
	MessageSender->HighestNum = receivedSN;

      } else {

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Demon::ReceiveData(%u) : REQUESTING PACKETS 1 to %u from %u we %u\n",
		toHandle, receivedSN, MessageSender->SSRC, c.We[toHandle]->SSRC);
#endif

	for (unsigned int x = 1; x < receivedSN; x++) {

#ifdef _SMP_REQTIMER
	  fprintf(SMP::debug, "Set first request timer for %u toHandle = %u delay = %u\n",
		  x, toHandle, MessageSender->delay);
#endif

	  // Timer stellen fuer Packet x
	  SetTimer(toHandle, receivedSSRC,
		   x, Timer::REQUEST, MessageSender->delay,
		   MessageSender, c.We[toHandle]->SSRC,
		   0, true, 0, 0, FIRSTTIME);

	}
      }
    }
    // *******************************************************************
    // no packets are missing => highest received sn = received sn
    // *******************************************************************
    else {
      MessageSender->HighestNum = receivedSN;
    }

    // *******************************************************************
    // insert packet in operation list of message sender
    // (for later requests by other members)
    // *******************************************************************
    if (InsertOpList(toHandle, MessageSender, data, packetSize,
		     receivedSN) == NULL) {
      fprintf(SMP::debug, "Demon::receiveData() : could not insert op\n");
    }

#ifdef _SMP_SHOWLIST
    timerlist.Print();
#endif

    if ((receivedSN > 1) && (c.LateJoin[toHandle])) {

      if (c.SourceOrdering[toHandle]) {
	ReceivedADU* adu = new ReceivedADU(receivedSN, data, bread);
	if (!adu) {
	  fprintf(SMP::debug, "Demon::receive - operation new ReceivedADU failed\n");
	  fprintf(SMP::debug, "socket is being closed - try to connect again\n");
	  logOff(toHandle);
	  Demon::close(toHandle);
	  return(-1);
	}
	MessageSender->receivedlist.Insert(adu);

#ifdef _SMP_LJ
	fprintf(SMP::debug, "Client %u RSN > ESN : INSERT ReceiveList %u\n", toHandle, receivedSN);
#endif
      }
    } else {
      IndData(toHandle, data, bread, orgIP);
    }
    return(bread);
  }


  // *******************************************************************
  // sender known (already received a packet or a session message)
  // *******************************************************************
  unsigned int ESN = MessageSender->SN + 1;


  // *******************************************************************
  // received sequ.nr. equals expected seq.nr.
  // *******************************************************************
  if (receivedSN == ESN) {

#ifdef _SMP_SORDER
    fprintf(SMP::debug, "Client %u RSN = ESN : %u = %u\n", toHandle, receivedSN, ESN);
#endif

    // Packet in richtiger Reihenfolge eingetroffen
    // 1. Update von der Sequencenummer des Senders
    // 2. Hoechste ausgelieferte Nummer updaten
    // 3. Ueperpruefen, ob nicht ein REQ-Timer gestellt wurde
    // 4. Benachrichtige Anwendung
    MessageSender->SN = ESN;

    if ((MessageSender->HighestNum + 1) == receivedSN)
      MessageSender->HighestNum = ESN;
    MessageSender->timestampSent = receivedTS;
    MessageSender->timestampReceived = GetTimestamp();

    //MessageSender->active = true;

    if (!(c.MSet[toHandle]->Find(MessageSender)))
      c.MSet[toHandle]->Insert(MessageSender);

    // *******************************************************************
    // insert packet in operations list of sender for later requests
    // *******************************************************************
    if (InsertOpList(toHandle, MessageSender, data, bread,
		     receivedSN) == NULL) {
      fprintf(SMP::debug, "Demon::receiveData() : could not insert op\n");
    }

    if (c.SourceOrdering[toHandle] &&
	(receivedSN > MessageSender->HighestNum + 1)) {

      ReceivedADU* adu = new ReceivedADU(receivedSN, data, bread);

      if (!adu) {
	fprintf(SMP::debug, "Demon::receive - operation new ReceivedADU failed\n");
	fprintf(SMP::debug, "socket is being closed - try to connect again\n");
	logOff(toHandle);
	Demon::close(toHandle);
	return(-1);
      }
      MessageSender->receivedlist.Insert(adu);

#ifdef _SMP_SORDER
      fprintf(SMP::debug, "Client %u RSN > HSN+1 => INSERT ReceiveList: %u\n", toHandle, receivedSN);
#endif

    } else {

#ifdef _SMP_SORDER
      fprintf(SMP::debug, "Client %u RSN = HSN+1 => deliver: %u\n", toHandle, receivedSN);
#endif

      IndData(toHandle, data, bread, orgIP);
    }

    // *******************************************************************
    // check if a request timer has been set for this packet
    // *******************************************************************
    Timer* tt = FindTimer(toHandle, Timer::REQUEST, receivedSSRC, receivedSN);
    if (tt) {

#ifdef _SMP_TIMER
      fprintf(SMP::debug, "delete timer #1\n");
#endif
      tt =  DeleteTimer(toHandle, Timer::REQUEST, receivedSSRC, receivedSN);

      delete tt;
    }
  }

  // *******************************************************************
  // received sequ.nr. greater than expected seq.nr.
  // => gap in sequence numbers => request missing packets
  // *******************************************************************
  else if (receivedSN > ESN) {

#ifdef _SMP_SORDER
    fprintf(SMP::debug, "Client %u RSN > ESN : %u > %u\n", toHandle, receivedSN, ESN);
#endif

#ifdef _SMP_REQTIMER
    fprintf(SMP::debug, "Demon::ReceiveData(%u) - 2 : REQUESTING PACKETS %u to %u from %u we %u\n",
	    toHandle, ESN, receivedSN, MessageSender->SSRC, c.We[toHandle]->SSRC);
#endif

    for (unsigned int i = ESN; i < receivedSN; i++) {

      Timer* forTimer = FindTimer(toHandle, Timer::REQUEST, receivedSSRC, i);

      if (forTimer == NULL) {

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Set first request timer for %u toHandle %u delay %u\n",
		i, toHandle, MessageSender->delay);
#endif

	SetTimer(toHandle, receivedSSRC, i, Timer::REQUEST,
		 MessageSender->delay, MessageSender,
		 c.We[toHandle]->SSRC,0,true,
		 0, 0, FIRSTTIME);
      }
    }
    MessageSender->SN = receivedSN;
    MessageSender->timestampSent = receivedTS;
    MessageSender->timestampReceived = GetTimestamp();

    // MessageSender->active = true;
    if (!(c.MSet[toHandle]->Find(MessageSender)))
      c.MSet[toHandle]->Insert(MessageSender);

    // *******************************************************************
    // insert packet in operations list of sender for later requests
    // *******************************************************************
    if (InsertOpList(toHandle, MessageSender, data, bread,
		     receivedSN) == NULL) {
      fprintf(SMP::debug, "Demon::receiveData() : could not insert op\n");
    }

    if (c.SourceOrdering[toHandle]) {
      ReceivedADU* adu = new ReceivedADU(receivedSN, data, bread);
      if (!adu) {
	fprintf(SMP::debug, "Demon::receive - operation new ReceivedADU failed\n");
	fprintf(SMP::debug, "socket is being closed - try to connect again\n");
	logOff(toHandle);
	Demon::close(toHandle);
	return(-1);
      }
      MessageSender->receivedlist.Insert(adu);

#ifdef _SMP_SORDER
      fprintf(SMP::debug, "Client %u RSN > ESN : INSERT ReceiveList %u\n", toHandle, receivedSN);
#endif

    } else {
      IndData(toHandle, data, bread, orgIP);
    }
  } // E2

  // *******************************************************************
  // received sequ.nr. less than expected seq.nr.
  // - received duplicate OR
  // - received repair
  // *******************************************************************
  else if (receivedSN < ESN) {

#ifdef _SMP_SORDER
    fprintf(SMP::debug, "Client %u RSN < ESN : %u < %u\n", toHandle, receivedSN, ESN);
#endif

    // *****************************************************************
    // the received seq.nr. is less than the seq.nr. already delivered
    // to the application. Thus, we received a real duplicate
    // *****************************************************************
    if (receivedSN <= MessageSender->HighestNum) {

#ifdef _SMP_SORDER
      fprintf(SMP::debug, "Client %u RSN <= HSN : %u <= %u => duplicate\n",
	      toHandle, receivedSN, MessageSender->HighestNum);
#endif
    }

    // *****************************************************************
    // the received seq.nr. is greater than the highest number delivered
    // to the application but smaller than expected number
    // *****************************************************************
    else {

#ifdef _SMP_SORDER
      fprintf(SMP::debug, "Client %u RSN > HSN : %u > %u\n",
	      toHandle, receivedSN, MessageSender->HighestNum);
#endif

      // *******************************************************************
      // check receive list, packet already received but not delivered
      // => duplicate, free buffer
      // *******************************************************************
      ReceivedADU adu(receivedSN, data, bread);
      ReceivedADU* aduP = MessageSender->receivedlist.Find(&adu);
      if (aduP != NULL) {

#ifdef _SMP_SORDER
	fprintf(SMP::debug, "Demon::ReceiveData() : received duplicate\n");;
#endif

	FreePacketBuffer(toHandle, data, bread);
      }
      // *******************************************************************
      // packet not received yet => deliver or put in source ordering list
      // or deliver if source ordering is disabled
      // *******************************************************************
      else {

	// *****************************************************************
	// check if we have already set a request timer for the packet
	// if not, an error occured somewhere. The request timer should
	// always be set for THIS packet
	// *****************************************************************
	Timer* timer = FindTimer(toHandle, Timer::REQUEST,
				 receivedSSRC, receivedSN);
	if (timer == NULL) {
	  fprintf(SMP::debug, "Demon::receive - No REQ Timer set\n");
	} else {
	  Timer* ExpiredTimer = timer;
	  if (timer->ActionId == Timer::REQUEST) {

#ifdef _SMP_REQTIMER
	    fprintf(SMP::debug, "Demon::ReceiveData() : request timer exists => delete\n");
#endif

#ifdef _SMP_TIMER
	   fprintf(SMP::debug, "delete timer #2\n");
#endif

	    Timer * tp = DeleteTimer(toHandle, Timer::REQUEST,
				     receivedSSRC, receivedSN);
	    delete tp;
	  }
	}

	// *******************************************************************
	// insert packet in operations list of sender for later requests
	// *******************************************************************
	Operation *op;
	if ((op = InsertOpList(toHandle, MessageSender, data, bread,
			       receivedSN)) == NULL) {
	  fprintf(SMP::debug, "Demon::receiveData() : could not insert op\n");
	} else {
	  op->repair_time = GetTimestamp();
	}

	// ****************************************************************
	// seq.nr. greater than required number for delivery
	// => deliver if source ordering disabled otherwise put in receive
	//    buffer
	// ****************************************************************
	if (receivedSN > ((MessageSender->HighestNum) + 1)) {

#ifdef _SMP_SORDER
	  fprintf(SMP::debug, "Client %u RSN > HSN + 1 : %u = %u insert recv list or deliver\n",
		  toHandle, receivedSN, MessageSender->HighestNum + 1);
#endif
	  if (c.SourceOrdering[toHandle]) {
	    ReceivedADU* adu = new ReceivedADU(receivedSN, data, bread);
	    if (!adu) {
	      fprintf(SMP::debug, "Demon::receive - operation new ReceivedADU failed\n");
	      fprintf(SMP::debug, "socket is being closed - try to connect again\n");
	      logOff(toHandle);
	      Demon::close(toHandle);
	      return(-1);
	    }
	    MessageSender->receivedlist.Insert(adu);

#ifdef _SMP_SORDER
	    fprintf(SMP::debug, "Client %u RSN > HSN + 1 : INSERT ReceiveList: %u\n", toHandle, receivedSN);;
#endif

	  } else {
	    IndData(toHandle, data, bread, orgIP);
	  }
	}

	// ****************************************************************
	// seq.nr. equals required number for delivery (highest + 1)
	// => packet number + subsequent complete packet numbers in
	// receive buffer are due for delivery
	// ****************************************************************
	else {

#ifdef _SMP_SORDER
	  fprintf(SMP::debug, "Client %u RSN = HSN + 1 : %u = %u deliver and check recv list\n",
		  toHandle, receivedSN, MessageSender->HighestNum + 1);
#endif

	  MessageSender->HighestNum++;
	  IndData(toHandle, data, bread, orgIP);

	  if (c.SourceOrdering[toHandle]) {
	    ReceivedADU* ToDeliver = MessageSender->receivedlist.First();
	    while ((ToDeliver) &&
		   (ToDeliver->SN == (MessageSender->HighestNum + 1))) {

#ifdef _SMP_SORDER
	      fprintf(SMP::debug, "Client %u : SEQ.NR. of delivered p. = %u\n", toHandle,ToDeliver->SN);
#endif

	      IndData(toHandle, ToDeliver->data, ToDeliver->size,
		      orgIP);
	      delete(MessageSender->receivedlist.RemoveFirst());
	      MessageSender->HighestNum++;
	      ToDeliver = MessageSender->receivedlist.First();
	    }
	  }
	}
      }
    }
  }
  return(bread);
}


/*********************************************************************
 Name:      Demon::ReceiveUnreliableData()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveUnreliableData(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  //fprintf(SMP::debug, "Demon::ReceiveUnreliableData - for client %u\n", toHandle);

  bool senderUnKnown = false;

  // *******************************************************************
  // read header data
  // *******************************************************************
  SMPHeader* Header = (SMPHeader*)data;

  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;
  unsigned int orgIP = Header->ORGIP;

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {
    //if (c.Loopback[toHandle]) { // client wishes to receive his own packets
    //IndData(toHandle, data, bread, orgIP);
    //FreePacketBuffer(toHandle, data, bread);
    //return(0);
    //} else {
    FreePacketBuffer(toHandle, data, bread);
    return(0);
    //}
  }

  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

  if (!receivedX)
    Header->FLAGS = htons(0 << 15 | ntohs(Header->FLAGS)&SMP_PT);

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Demon::ReceiveUnreliableData - Client[%u]: receive from Member %u size %u\n",
	  toHandle, receivedSSRC, bread);
#endif

#ifdef _SMP_UNRELIABLE
  fprintf(SMP::debug, "Demon::SendUnreliableData - received %u bytzes\n", bread);
#endif

  // *******************************************************************
  // check member (check must not be interrupted!)
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);
  if (!MessageSender) {
    FreePacketBuffer(toHandle, data, bread);
    fprintf(SMP::debug, "Demon::receive(%u) : member error\n");
    return(-1);
  }

  IndUnreliableData(toHandle, data, bread, orgIP);

  return(bread);
}


/*********************************************************************
 Name:      Demon::ReceiveJoin()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveJoin(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;

  // *******************************************************************
  // read header data
  // *******************************************************************

  SMPHeader* Header = (SMPHeader*)data;
  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {

    FreePacketBuffer(toHandle, data, bread);

    return(0);
  }
  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Client[%u]: receive at position %u bytes %u\n", toHandle, Header, bread);
#endif

  // *******************************************************************
  // check member
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);

  if (MessageSender == NULL) {
    fprintf(SMP::debug, "Demon::ReceiveJoin(%u) : could not create new member %u\n", receivedSSRC);
    FreePacketBuffer(toHandle, data, bread);

    return(-1);
  } else
    FreePacketBuffer(toHandle, data, bread);

  return(bread);
}

/*********************************************************************
 Name:
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveLeave(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;

  // *******************************************************************
  // read header data
  // *******************************************************************

  SMPHeader* Header = (SMPHeader*)data;
  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {

    FreePacketBuffer(toHandle, data, bread);

    return(0);
  }


  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Client[%u]: receive at position %u bytes %u\n", toHandle, Header, bread);
#endif

  // *******************************************************************
  // check member
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);

  c.Members[toHandle]--;
  if (!MessageSender) {
    fprintf(SMP::debug, "LEAV: Leaving Application unkown\n");
    FreePacketBuffer(toHandle, data, bread);

    return(-1);
  }
  MessageSender->ActiveMember = false;


  if (!c.LateJoin[toHandle]) {
    UpdateAllOpLists(toHandle, receivedSSRC, 0);
  }
  FreePacketBuffer(toHandle, data, bread);

  return(bread);
}

/*********************************************************************
 Name:
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveRequest(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;

  // *******************************************************************
  // read header data
  // *******************************************************************
  SMPHeader* Header = (SMPHeader*)data;

  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Demon::ReceiveRequest - packet from (%u,%u,%u,%u): packetSize %u\n",
	  receivedSSRC, receivedSN, receivedTS, receivedPT, packetSize);
#endif

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {
    FreePacketBuffer(toHandle, data, bread);
    return(0);
  }

  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Client[%u]: receive at position %u bytes %u\n", toHandle, Header, bread);
#endif

  // *******************************************************************
  // check member
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);

  // *******************************************************************
  // sender not known => we can not handle the request
  // *******************************************************************
  if (!MessageSender) {
#ifdef _SMP_REQTIMER
    fprintf(SMP::debug, "Demon::ReceiveRequest() : sender unknown ... ignore\n");
#endif
    FreePacketBuffer(toHandle, data, bread);

    return(-1);
  }

  // *******************************************************************
  // check request body: if original sender unknown => return
  // *******************************************************************
  SMPRequestBody* RequestBody = (SMPRequestBody *)
    (data + sizeof(SMPHeader));


  Member m(ntohl(RequestBody->ReqSSRC));
  Member* OriginalSender = c.grouplist[toHandle].Find(&m);

  if (OriginalSender == NULL) {

#ifdef _SMP_DEBUG
    fprintf(SMP::debug, "REQ: original Sender with SSRC %u unkown\n", ntohl(RequestBody->ReqSSRC));
#endif

    FreePacketBuffer(toHandle, data, bread);

    return(-1);
  }

#ifdef _SMP_REQTIMER
  fprintf(SMP::debug, "Demon::ReceiveRequest() : %u requests %u/%u\n",
	  receivedSSRC, OriginalSender->SSRC, ntohl(RequestBody->ReqSN));
#endif

  // *******************************************************************
  // look if we have the requested data:
  // 1. the SN number of the original sender must be higher than the
  //    requested number and
  // 2. data must be in operations list
  // if we have the data set repair timer and reset request timer
  // if we don't have the data we request the data
  // *******************************************************************
  if (ntohl(RequestBody->ReqSN) <= OriginalSender->SN) {

    Operation toFind(ntohl(RequestBody->ReqSN));
    Operation *op = OriginalSender->operationslist.Find(&toFind);

    // *******************************************************************
    // we have the data => set repair timer
    // *******************************************************************
    if (op) {

      // *******************************************************************
      // check if we just received or sent a repair for that packet
      // if yes don't set a request timer
      // *******************************************************************
      if ((op->repair_time != 0) &&
	  ((GetTimestamp() - op->repair_time) < 3*MessageSender->delay)) {

      } else {

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Set repair timer for %u client %u delay = %u\n",
		ntohl(RequestBody->ReqSN), toHandle , MessageSender->delay);
#endif

	SetTimer(toHandle, ntohl(RequestBody->ReqSSRC),
		 ntohl(RequestBody->ReqSN), Timer::REPAIR,
		 MessageSender->delay,
		 OriginalSender,
		 c.We[toHandle]->SSRC, 0, false,
		 (unsigned char) ntohs(RequestBody->ReqTTL),
		 false, FIRSTTIME);
      }
    }

    // *******************************************************************
    // we don't have the data: check if request timer exists
    // *******************************************************************
    else {

      Timer* T = FindTimer(toHandle, Timer::REQUEST,
			   (unsigned int)ntohl(RequestBody->ReqSSRC),
			   (unsigned int)ntohl(RequestBody->ReqSN));
      if (T) {

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Request timer already set for %u / %u => reset timer\n", T->SSRC, T->SN);
#endif

#ifdef _SMP_SHOWLIST
	timerlist.Print();
#endif
	if (T->ActionId == Timer::REQUEST) {

#ifdef _SMP_TIMER
	  fprintf(SMP::debug, "delete timer #3\n");
#endif

	  Timer* t =  DeleteTimer(toHandle, Timer::REQUEST,
				  (unsigned int)ntohl(RequestBody->ReqSSRC),
				  (unsigned int)ntohl(RequestBody->ReqSN));
#ifdef _SMP_REQTIMER
	  fprintf(SMP::debug, "Demon::ReceiveRequest() : new request timer for %u toHandle %u delay %u\n",
		  ntohl(RequestBody->ReqSN), toHandle, MessageSender->delay);
#endif

	  SetTimer(toHandle, ntohl(RequestBody->ReqSSRC),
		   ntohl(RequestBody->ReqSN), Timer::REQUEST,
		   OriginalSender->delay, OriginalSender,
		   c.We[toHandle]->SSRC, 0, false,0,
		   true, (t->i + 1));

	  delete t;
	}
      } else {
	fprintf(SMP::debug, "Demon::ReceiveRequest(%u) : error, no request timer set for %u / %u\n",
		toHandle, ntohl(RequestBody->ReqSSRC), ntohl(RequestBody->ReqSN));
      }
    }
  }

  // *******************************************************************
  // somebody is requesting data which we don't have => request
  // *******************************************************************
  else {

    unsigned int FirstSequenceNumber = OriginalSender->SN + 1;
    unsigned int LastSequenceNumber = ntohl(RequestBody->ReqSN) + 1;

#ifdef _SMP_REQTIMER
    fprintf(SMP::debug, "Demon::ReceiveRequest(%u) - 2 : REQUESTING PACKETS %u to %u from %u we %u\n",
	    toHandle, FirstSequenceNumber, LastSequenceNumber, ntohl(RequestBody->ReqSSRC), c.We[toHandle]->SSRC);
#endif

    for (unsigned int Req = FirstSequenceNumber;
	 Req < LastSequenceNumber; Req++) {

#ifdef _SMP_REQTIMER
      fprintf(SMP::debug, "Set request timer for %u client: %u delay %u\n", Req, toHandle, MessageSender->delay);
#endif
      Timer* forTimer = FindTimer(toHandle, Timer::REQUEST,
				  ntohl(RequestBody->ReqSSRC), Req);

      if (forTimer == NULL) {

	SetTimer(toHandle, ntohl(RequestBody->ReqSSRC),
		 Req, Timer::REQUEST,
		 OriginalSender->delay,
		 OriginalSender,
		 c.We[toHandle]->SSRC, 0,
		 true, 0, true, FIRSTTIME);
      }
    }
  }

  FreePacketBuffer(toHandle, data, bread);

  return(bread);
}

/*********************************************************************
 Name:      Demon::ReceiveSessionMsg()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveSessionMsg(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;
  unsigned int OldSN;

  // *******************************************************************
  // read header data
  // *******************************************************************
  SMPHeader* Header = (SMPHeader*)data;

  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {
    FreePacketBuffer(toHandle, data, bread);
    return(0);
  }

  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Client[%u]: receive at position %u bytes %u\n", Header, bread, bytes);
#endif

  // *******************************************************************
  // check member (this must not be interrupted by a signal otherwise
  // there may be inconsistencies!)
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);

  if (!MessageSender) {
    fprintf(SMP::debug, "Demon::ReceiveSessionMsg() : error with message sender\n");
    FreePacketBuffer(toHandle, data, bread);

    return(-1);
  }

  if (MessageSender->SN < receivedSN) {
    OldSN = MessageSender->SN;
    MessageSender->SN = receivedSN;
  } else {
    OldSN = receivedSN;
  }

  MessageSender->timestampSent = receivedTS;
  u_int32 time = GetTimestamp();
  MessageSender->timestampReceived = time;
  MessageSender->LastSessionMessage = time;

  // *******************************************************************
  // unknown sender => request missing data or send LJIND
  // *******************************************************************
  if (senderUnKnown) {

#ifdef _SMP_SESS
    fprintf(SMP::debug, "Demon::receive() : received sm, size %u _ sender unknown %u deq no %u\n",
	    packetSize, toHandle, receivedSN);
#endif

    if (!c.LateJoin[toHandle]) {

      if (receivedSN > 0) {
	if (c.FirstLateJoinInd[toHandle]) {

#ifdef _SMP_LJ
	  fprintf(SMP::debug, "Demon::receive() : received sessionmessage: send LJIND to app to Handle %u\n", toHandle);
#endif

	  IndLateJoin(toHandle, receivedSSRC);
	  c.FirstLateJoinInd[toHandle] = false;
	}
      }
      // regulaeren Repair-Mechanismus unterdruecken
      MessageSender->HighestNum = receivedSN;

    } else {

      for (unsigned int x = 1; x <= receivedSN; x++) {

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Demon::ReceiveSessionMsg() : new req timer for %u toHandle %u delay %u\n",
		x, toHandle, MessageSender->delay);
#endif
	Timer* forTimer = FindTimer(toHandle, Timer::REQUEST, receivedSSRC, x);

	if (forTimer == NULL) {

	  SetTimer(toHandle, receivedSSRC,
		   x, Timer::REQUEST, MessageSender->delay,
		   MessageSender, c.We[toHandle]->SSRC,
		   0, true, 0, 0, FIRSTTIME);
	}

      }
    }
  }

  // *******************************************************************
  // known sender => check already received packets
  // *******************************************************************
  else {

#ifdef _SMP_SESS
    fprintf(SMP::debug, "Demon::receive() : received sm size << %u : sender known %u seq no %u\n",
	    packetSize, toHandle, receivedSN);
#endif

    MessageSender->LastSessionMessage = GetTimestamp();

    if (OldSN < receivedSN) {

      unsigned int ExpectedSequenceNumber = OldSN + 1;
      for (unsigned int j = ExpectedSequenceNumber; j <= receivedSN; j++) {

#ifdef _SMP_REQTIMER
	fprintf(SMP::debug, "Demon::ReceiveSessionMsg() : new req timer for %u toHandle %u delay %u\n",
		j, toHandle, MessageSender->delay);
#endif

	Timer* forTimer = FindTimer(toHandle, Timer::REQUEST, receivedSSRC, j);
	if (forTimer == NULL) {

	  SetTimer(toHandle, receivedSSRC,
		   j, Timer::REQUEST, MessageSender->delay,
		   MessageSender, c.We[toHandle]->SSRC,
		   0, true, 0, false, FIRSTTIME);
	}
      }
    }
  }

  // Sender der Session Message bearbeitet
  // angefuehrte Mitglieder auswerten
  ComputeSessionMessage(toHandle, data, receivedSSRC, receivedTS,
			MessageSender);


  FreePacketBuffer(toHandle, data, bread);

  return(bread);
}

/*********************************************************************
 Name:      Demon::ReceiveDADUReq()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveDADUReq(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;

  // *******************************************************************
  // read header data
  // *******************************************************************
  SMPHeader* Header = (SMPHeader*)data;

  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;
  unsigned int orgIP = Header->ORGIP;

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {
    FreePacketBuffer(toHandle, data, bread);
    return(0);
  }

  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

  if (!receivedX)
    Header->FLAGS = htons(0 << 15 | ntohs(Header->FLAGS)&SMP_PT);

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Client[%u]: receive at position %u bytes %u\n", data, bread);
#endif

  // *******************************************************************
  // check member
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);

  if (!MessageSender) {
    FreePacketBuffer(toHandle, data, bread);

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Demon::ReceiveDADUReq(%u) : member does not exist\n");
#endif

    FreePacketBuffer(toHandle, data, bread);

    return(-1);
  }

  // nur antworten, wenn es nicht der eigene request ist
  if ((receivedSSRC != c.We[toHandle]->SSRC)) {

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Client[%u] received DelayedADU-Request\n", toHandle);
#endif

    DADU* dadu;

    // *******************************************************************
    // check if request is already being processed (REPEAT transmission!)
    // *******************************************************************
    DADU toFind(receivedSSRC, receivedSN);
    dadu = c.dadulist[toHandle].Find(&toFind);
    if (dadu) {
#ifdef _SMP_LJ
      fprintf(SMP::debug, "Request is already being processed (duplicate request)\n");
#endif

      if ((dadu->reply_time != 0) &&
	  ((GetTimestamp() - dadu->reply_time) < 3*MessageSender->delay)) {
#ifdef _SMP_LJ
	fprintf(SMP::debug, "... old request not timed out => ignoring request\n");
#endif
	FreePacketBuffer(toHandle, data, bread);
	return(-1);
      } else if (dadu->reply_time != 0) {
#ifdef _SMP_LJ
	fprintf(SMP::debug, "... deleting old request, creating new one (time-out)\n");
#endif
	DADU * toDelete = c.dadulist[toHandle].Delete(dadu);
	if (toDelete)
	  delete(toDelete);
	else
	  fprintf(SMP::debug, "Demon::SendDADURsp() : could not delete DADU request\n");
      } else {
#ifdef _SMP_LJ
	fprintf(SMP::debug, "... runing givethrough timer ... ignoring request\n");
#endif
	FreePacketBuffer(toHandle, data, bread);
	return(-1);
      }
    }

    // *******************************************************************
    // check if timer has been already set
    // *******************************************************************
    //Timer* gtTimer = FindTimer(toHandle, Timer::GIVETHROUGH, receivedSSRC,
    //			       receivedSN);
    //if (gtTimer != NULL) {
    //#ifdef _SMP_LJ
    //      fprintf(SMP::debug, "Givethrough-Timer was already set\n");
    //#endif
    //      return(-1);
    //    }

    // *******************************************************************
    // insert dadu request in DADU list
    // *******************************************************************
    dadu = new DADU(receivedSSRC, receivedSN, data, bread, orgIP);
    c.dadulist[toHandle].Insert(dadu);

    SetTimer(toHandle, receivedSSRC,
	     receivedSN, Timer::GIVETHROUGH, MessageSender->delay,
	     0, c.We[toHandle]->SSRC,
	     0, true, 0, 0, FIRSTTIME);

    dadu->status = GT_TIMER_RUNNING;

#ifdef _SMP_SHOWLIST
    timerlist.Print();
#endif
  }

  FreePacketBuffer(toHandle, data, bread);

  return(bread);
}

/*********************************************************************
 Name:      Demon::ReceiveDADURsp()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
int SMP::Demon::ReceiveDADURsp(int toHandle, char* data, int bread, sockaddr_in fromaddr) {

  bool senderUnKnown = false;

  // *******************************************************************
  // read header data
  // *******************************************************************
  SMPHeader* Header = (SMPHeader*)data;

  int packetSize = ntohs(Header->SZ);
  unsigned int receivedSSRC = ntohl(Header->SSRC);
  int receivedPT = ntohs(Header->FLAGS)&SMP_PT;
  unsigned int receivedTS = ntohl(Header->TS);
  unsigned int receivedSN = ntohl(Header->SN);
  bool receivedX = ntohs(Header->FLAGS)&SMP_X;

  //  fprintf(SMP::debug, "ReceiveDADURsp() : DADUbread = %u size %u\n", bread, packetSize);

  // *******************************************************************
  // own packet => return
  // *******************************************************************
  if (c.We[toHandle]->SSRC == receivedSSRC) {
    FreePacketBuffer(toHandle, data, bread);
    return(0);
  }

  if (bread != ntohs(Header->SZ))
    fprintf(SMP::debug, ">>>>> ERROR : Header->SZ wrong <<<<<\n");

  // *******************************************************************
  // get sender ip address and port
  // *******************************************************************
  unsigned int receivedIpAddr = fromaddr.sin_addr.s_addr;
  unsigned short receivedPort = fromaddr.sin_port;

  if (!receivedX)
    Header->FLAGS = htons(0 << 15 | ntohs(Header->FLAGS)&SMP_PT);

#ifdef _SMP_DEBUG
  fprintf(SMP::debug, "Client[%u]: receive at position %u bytes %u\n", toHandle, data , bread);
#endif

  // *******************************************************************
  // check member
  // *******************************************************************
  Member* MessageSender = CheckMember(toHandle, receivedSSRC, receivedIpAddr,
				      receivedPort, receivedTS, receivedPT,
				      receivedX, &senderUnKnown);

  if (!MessageSender) {

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Demon::ReceiveDADURsp(%u) : member does not exist ... cont\n");
#endif

  }
  SMPDelADURespBody* delADUResp = (SMPDelADURespBody*)
    (data + sizeof(SMPHeader));

  // *******************************************************************
  // received a response for our own request
  // 1. delete REPEAT timer
  // 2. accept new request by the client (OpenRequest == false)
  // *******************************************************************
  if (ntohl(delADUResp->RSSRC) == c.We[toHandle]->SSRC) {

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Client[%u] received DelayedADU-Response for sended Request\n", toHandle);
#endif

#ifdef _SMP_TIMER
    fprintf(SMP::debug, "delete timer #4\n");
#endif

    Timer* tt = DeleteTimer(toHandle, Timer::REPEAT, c.We[toHandle]->SSRC, 0);
    if (tt) {
      delete tt;
    }

    IndDADURsp(toHandle, data, bread, receivedIpAddr);

    c.OpenRequest[toHandle] = false;
  }

  // *******************************************************************
  // received a response for a request of another member
  // 1. delete GIVETHROUGH timer
  // 2. delete DADU in list
  // *******************************************************************
  if (ntohl(delADUResp->RSSRC) != c.We[toHandle]->SSRC) {

#ifdef _SMP_LJ
    fprintf(SMP::debug, "Client[%u] received DelayedADU-Response of another member\n", toHandel);
#endif

    DADU toFind(ntohl(delADUResp->RSSRC), receivedSN);
    DADU * dadu = c.dadulist[toHandle].Find(&toFind);
    if (dadu) {

#ifdef _SMP_LJ
      fprintf(SMP::debug, "Demon::ReceiveDADURsp() : DADU found\n");
#endif

      if (dadu->status == GT_TIMER_RUNNING) {

#ifdef _SMP_LJ
	fprintf(SMP::debug, "Demon::ReceiveDADURsp() : GT timer is running ... delete timer and DADU\n");
#endif

#ifdef _SMP_TIMER
	fprintf(SMP::debug, "delete timer #5 \n");
#endif
	Timer* tt = DeleteTimer(toHandle, Timer::GIVETHROUGH,
				ntohl(delADUResp->RSSRC), receivedSN);
	if (tt) {
	  delete tt;
	}
	dadu = c.dadulist[toHandle].Delete(&toFind);
	if (dadu) delete(dadu);
      }
    } else {

#ifdef _SMP_LJ
      fprintf(SMP::debug, "Demon::ReceiveDADURsp() : could not find DADU\n");
#endif

    }

    Member ToFind(ntohl(delADUResp->RSSRC));
    Member* originMember = c.grouplist[toHandle].Find(&ToFind);
    if (originMember == NULL) {
      fprintf(SMP::debug, "Demon::receive - Receiver of DelayedADUResponse is not a Member\n");
      FreePacketBuffer(toHandle, data, bread);

      return(-1);
    }
  }

  FreePacketBuffer(toHandle, data, bread);


  return(bread);
}

/*********************************************************************
 Name:
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
char* SMP::Demon::AllocatePacketBuffer(int toHandle, int size) {
  char * data = (char*)malloc(sizeof(char)*size);
  return data;
}

/*********************************************************************
 Name:      Demon::FreePacketBuffer()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
void SMP::Demon::FreePacketBuffer(int toHandle, char * data, int size, bool block) {
  free(data);
}

/*********************************************************************
 Name:      Demon::InsertOpList()
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
SMP::Operation* SMP::Demon::InsertOpList(int toHandle, Member* m, void* data,
			       int size, unsigned int sn, bool block) {

  Operation* Op = new Operation(sn, data, size);
  if (!Op) {
    fprintf(SMP::debug, "Demon::InsertOpList() - operation new Operation failed\n");
    return(NULL);
  }

  Member* M = c.grouplist[toHandle].First();
  unsigned int WeSSRC = c.We[toHandle]->SSRC;
  while (M) {

    if ((M->ActiveMember) && (M->SSRC != WeSSRC) ) {
      Op->MemberList.insert(M->SSRC);
    };
    M = c.grouplist[toHandle].Next();
  }

  if (m != NULL) {
    m->operationslist.Insert(Op);
    Op->insert_time = GetTimestamp();
  } else {
    fprintf(SMP::debug, "Demon::InsertOpList() - could not find member %u\n ", m->SSRC);
    return(NULL);
  }

#ifdef _SMP_SHOWLIST
  c.We[toHandle]->operationslist.Print();
#endif
  return(Op);
}

/*********************************************************************
 Name:
 Param:
 Result:
 Desc:
 Calledby:
 Uses:
 *********************************************************************/
bool SMP::Demon::DeleteOpList(int toHandle, Member* m, unsigned int sn,
			bool block) {

  if (m) {
    Operation toFind(sn);
    Operation* op = m->operationslist.Find(&toFind);
    if (op) {
      FreePacketBuffer(toHandle, op->data, op->size, false);
      delete(m->operationslist.Delete(op));
    } else {
      fprintf(SMP::debug, "Demon::DeleteOpList() could not find op\n");
      return(false);
    }
  } else {
    fprintf(SMP::debug, "Demon::DeleteOpList() could not find member %u\n", m->SSRC);
    return(false);
  }
  return(true);
}

void SMP::Demon::UpdateOpList(int toHandle, Member* member, unsigned int ssrc,
		  unsigned int sn, bool block) {

  unsigned int now = GetTimestamp();

  if (!member) {
    fprintf(SMP::debug, "Demon::UpdateOpList() : NULL pointer\n");
    return;
  }

  Operation* op = member->operationslist.First();
  while((op) && ((op->OpId <= sn) || (sn == 0))) {

    op->MemberList.erase(ssrc);

    if (op->MemberList.empty() && ((now - op->creation_time) > OP_DISCARD)) {

#ifdef _SMP_ACK
      fprintf(SMP::debug, "Demon::UpdateOpList() : deleting %u of %u\n", op->OpId, member->SSRC);
#endif

      FreePacketBuffer(toHandle, op->data, op->size, false);
      delete(member->operationslist.Delete(op));

      op = member->operationslist.First();
    } else {
      op = member->operationslist.Next();
    }
  }
}

void SMP::Demon::UpdateAllOpLists(int toHandle, unsigned int ssrc,
			     unsigned int sn, bool block) {

  Member* m = c.grouplist[toHandle].First();
  while(m) {
    UpdateOpList(toHandle, m, ssrc, sn, block);
    m = c.grouplist[toHandle].Next();
  }
}
