// *******************************************************************
//
//         A Video Server Architekture Based on Server Arrays
//
//                            Project:
//
//       Prototype Implementation of a Client-Server-VOD System
//
//                 Johannes Dengler and Werner Geyer
//                        Institut EURECOM
//                     2229, Route des Cretes
//	                  Sophia Antipolis
//                         F-06560 France
//
//                        geyer@eurecom.fr
//                       dengler@eurecom.fr
//
// *******************************************************************
// file    : include/list.C
// authors : Laurent Gautier / Johannes Dengler
// date    : 01/09/1994
// update  : 29/09/1994
// desc    : contains generic classes LIST and SORTLIST that were very
//           useful in our software.
// *******************************************************************

#ifndef __LIST_H
#define __LIST_H

#include <stddef.h>

/*******************\
 * Liste generale  *
 *-----------------*
 * Dbl linked List *
 \*******************/
template <class ITEM> class LIST;

// Item de la liste
template <class ITEM>
class LISTITEM
{
public:
  LISTITEM(ITEM *ListItem)
    { Item=ListItem; Next=NULL; Prev=NULL; }

  ITEM *Item;           // Item de la list
  LISTITEM<ITEM> *Next; // Item suivant
  LISTITEM<ITEM> *Prev; // Item precedent

friend class LIST<ITEM>;
};

// Manager de la liste
template <class ITEM>
class LIST
{
public:
  LIST(void) { Init(); }
  ~LIST();

  // Initialisation: ListVide
  void Init(void) { Front=NULL; Back=NULL; Walk = NULL;}
  // Destruction d une liste et de ses items
  void Destroy(void);

  // Teste si la liste est vide
  int IsEmpty(void) { return(!Front); }

  // Ajout en debut de liste
  ITEM *AddFirst(ITEM *Item);
  // Ajout en fin de liste
  ITEM *AddLast(ITEM *Item);
  // Retrait en debut de liste
  ITEM *RemoveFirst(void);
  // Retrait en fin de liste
  ITEM *RemoveLast(void);
  // Consultation en debut de liste
  ITEM *First(void);
  // Consultation en fin de liste
  ITEM *Last(void);

  // added by J. Dengler: Get the next ITEM from the current pos.
  ITEM *Next(void);
  ITEM *Next(LISTITEM<ITEM> ** );
  ITEM *First(LISTITEM<ITEM> ** );
  // Consultation en fin de liste
  ITEM *Last(LISTITEM<ITEM> ** );

  // Concatene 2 list
  LIST &operator+=(LIST &List2);

protected:
  LISTITEM<ITEM> *Front;
  LISTITEM<ITEM> *Back;
  LISTITEM<ITEM> *Walk; // Walking pointer used for Next()

  // For some unknown reasons, an instaciated template losts
  //   her friend properties when it is inherited!!!
  // So I define here the usefull functions on LISTITEM, which
  //   access the non-public members of LISTITEM thanks to the friend
  //   property.
  ITEM *GetItem(LISTITEM<ITEM> *ListItem)
    { return(ListItem->Item); }
  void SetItem(LISTITEM<ITEM> *ListItem, ITEM *NewItem)
    { ListItem->Item=NewItem; }
  LISTITEM<ITEM> *GetNext(LISTITEM<ITEM> *ListItem)
    { return(ListItem->Next); }
  void SetNext(LISTITEM<ITEM> *ListItem, LISTITEM<ITEM> *NewNext)
    { ListItem->Next=NewNext; }
  LISTITEM<ITEM> *GetPrev(LISTITEM<ITEM> *ListItem)
    { return(ListItem->Prev); }
  void SetPrev(LISTITEM<ITEM> *ListItem, LISTITEM<ITEM> *NewPrev)
    { ListItem->Prev=NewPrev; }
};


/*****************\
 *  Liste triee  *
 *---------------*
 * Sortable List *
 \*****************/
template <class ITEM>
class SORTLIST : public LIST<ITEM>
{
public:
  SORTLIST(void) { }
  ~SORTLIST() {}

  void Print();
  void Insert(ITEM *Item);
  ITEM* Delete(ITEM *Item);
  void  DeleteALL(ITEM *Item);
  ITEM* Find(ITEM *Item);
};




// *******************************************************************
//
//         A Video Server Architekture Based on Server Arrays
//
//                            Project:
//
//       Prototype Implementation of a Client-Server-VOD System
//
//                 Johannes Dengler and Werner Geyer
//                        Institut EURECOM
//                     2229, Route des Cretes
//	                  Sophia Antipolis
//                         F-06560 France
//
//                        geyer@eurecom.fr
//                       dengler@eurecom.fr
//
// *******************************************************************
// file    : include/list.h
// authors : Laurent Gautier / Johannes Dengler
// date    : 01/09/1994
// update  : 29/09/1994
// desc    : contains generic classes LIST and SORTLIST that were very
//           useful in our software.
// *******************************************************************


/******************\
 * Liste generale *
\******************/
// Destruction de la liste et des items
template <class ITEM>
LIST<ITEM>::~LIST()
{
  Destroy();
}

template <class ITEM>
void LIST<ITEM>::Destroy(void)
{
  while (!IsEmpty())
    delete RemoveFirst();
  Walk = NULL;
}

// Ajout en debut de liste
template <class ITEM>
ITEM *LIST<ITEM>::AddFirst(ITEM *Item)
{
  LISTITEM<ITEM> *NewItem;

  NewItem=new LISTITEM<ITEM>(Item);
  SetPrev(NewItem, NULL);
  SetNext(NewItem, Front);
  if (IsEmpty())
    Back=NewItem;
  else
    SetPrev(Front, NewItem);
  Front=NewItem;
  return(Item);
}

// Ajout en fin de liste
template <class ITEM>
ITEM *LIST<ITEM>::AddLast(ITEM *Item)
{

  LISTITEM<ITEM> *NewItem;

  NewItem=new LISTITEM<ITEM>(Item);
  SetNext(NewItem, NULL);
  SetPrev(NewItem, Back);
  if (IsEmpty())
    Front=NewItem;
  else
    SetNext(Back, NewItem);
  Back=NewItem;
  return(Item);
}

// Retrait en debut de liste
template <class ITEM>
ITEM *LIST<ITEM>::RemoveFirst(void)
{
  ITEM *Ret;
  LISTITEM<ITEM> *pItem;

  if (IsEmpty())
    return(NULL);
  pItem=Front;
  if (Walk==Front)
    Walk=Front=GetNext(Front);
  else
    Front=GetNext(Front);
  if (Front==NULL) // ->IsEmpty()
    Back=NULL;
  else
    SetPrev(Front, NULL);
  Ret=GetItem(pItem);
  delete pItem;
  return(Ret);
}

// Retrait en fin de liste
template <class ITEM>
ITEM *LIST<ITEM>::RemoveLast(void)
{
  ITEM *Ret;
  LISTITEM<ITEM> *pItem;

  if (IsEmpty())
    return(NULL);
  pItem=Back;
  if (Walk == Back)
    Walk=Back=GetPrev(Back);
  else
    Back=GetPrev(Back);
  if (Back==NULL) // ->IsEmpty()
    Front=NULL;
  else
    SetNext(Back, NULL);
  Ret=GetItem(pItem);
  delete pItem;
  return(Ret);
}

// Consultation en debut de liste
template <class ITEM>
ITEM *LIST<ITEM>::First(void)
{
  Walk = Front;
  if (IsEmpty())
    return(NULL);

  return(GetItem(Walk));
}

// Consultation en debut de liste
template <class ITEM>
ITEM *LIST<ITEM>::First(LISTITEM<ITEM> ** ptr)
{
  if (ptr == NULL)
    return NULL;
  *ptr = Front;
  if (IsEmpty())
    return(NULL);

  return(GetItem(Front));
}

// Consultation en fin de liste
template <class ITEM>
ITEM *LIST<ITEM>::Last(void)
{
  Walk = Back;
  if (IsEmpty())
    return(NULL);

  return(GetItem(Back));
}

// Consultation en fin de liste
template <class ITEM>
ITEM *LIST<ITEM>::Last(LISTITEM<ITEM> ** ptr)
{

  if (ptr == NULL)
    return NULL;
  *ptr = Back;
  if (IsEmpty())
    return(NULL);

  return(GetItem(Back));
}

// Concatene 2 list
template <class ITEM>
LIST<ITEM> &LIST<ITEM>::operator+=(LIST<ITEM> &List2)
{
if (List2.Front!=NULL)
    {
    if (Front==NULL)
        {
        // Empty list
        Front=List2.Front;
        }
    else
        {
        SetNext(Back, List2.Front);
        SetPrev(List2.Front, Back);
        }
    Back=List2.Back;
    }
Walk = Front;
return(*this);
}


//Give next element in list
// Retrait en fin de liste
template <class ITEM>
ITEM *LIST<ITEM>::Next(void)
{
  if (Walk != NULL) {
    Walk = GetNext(Walk);
    if (Walk != NULL)
      return (GetItem(Walk));
    else
      return NULL;
  }
  else
    return NULL;
}



//Give next element in list
// Retrait en fin de liste
template <class ITEM>
ITEM *LIST<ITEM>::Next(LISTITEM<ITEM> ** ptr)
{
  if (ptr == NULL)
    return NULL;
  if (*ptr != NULL) {
    *ptr = GetNext(*ptr);
    if (*ptr != NULL)
      return (GetItem(*ptr));
    else
      return NULL;
  }
  else
    return NULL;
}



/*****************\
 *  Liste triee  *
 *---------------*
 * Sortable List *
\*****************/


template <class ITEM>
void SORTLIST<ITEM>::Insert(ITEM *Item)
{
LISTITEM<ITEM> *NewItem;
LISTITEM<ITEM> *CurItem;
if (Item == NULL) return;

// Find where to put the new item
CurItem=Front;
while (   (CurItem != NULL)
       && (*GetItem(CurItem) < *Item))
    {
    CurItem=GetNext(CurItem);
    }
if (CurItem==NULL)
    AddLast(Item);
else
    {
    if (GetPrev(CurItem) == NULL)
        AddFirst(Item);
    else
        {
        // Insert the new item at the found place
        NewItem=new LISTITEM<ITEM>(Item);
        SetPrev(NewItem, GetPrev(CurItem));
        SetNext(NewItem, CurItem);
        SetNext(GetPrev(CurItem), NewItem);
        SetPrev(CurItem, NewItem);
        }
    }
}


template <class ITEM>
ITEM* SORTLIST<ITEM>::Delete(ITEM *Item) {
  LISTITEM<ITEM> *CurItem;
  ITEM *RetItem;

  if (Item == NULL) return NULL;

  CurItem=Front;
  
  while ((CurItem != NULL)
	 && (*GetItem(CurItem) != *Item)) /* mg:(*GetItem(CurItem) < *Item))*/
    {
      CurItem=GetNext(CurItem);
    }
  if (CurItem != NULL) {      /* mg: (*GetItem(CurItem) == *Item) */
 
    if (GetPrev(CurItem) == NULL)
      return RemoveFirst();
    else {
      if (GetNext(CurItem) == NULL)
	return RemoveLast();
      else {
	SetNext(GetPrev(CurItem), GetNext(CurItem));
	SetPrev(GetNext(CurItem), GetPrev(CurItem));
	return GetItem(CurItem);
      }
    }
  } else
    RetItem=NULL;
  return(RetItem);
}


template <class ITEM>
void SORTLIST<ITEM>::DeleteALL(ITEM *Item) {

  LISTITEM<ITEM> *CurItem;
  LISTITEM<ITEM> *CurItemTmp;

  if (Item == NULL) return;

  CurItem=Front;
  while (CurItem != NULL) {

    if ((*GetItem(CurItem)).IsEqual(*Item)) {
      if (GetPrev(CurItem) == NULL) {
	RemoveFirst();
	CurItem = Front;
      } else if (GetNext(CurItem) == NULL) {
	RemoveLast();
	CurItem = Back;
      } else {
	CurItemTmp = GetPrev(CurItem);
	SetNext(GetPrev(CurItem), GetNext(CurItem));
	SetPrev(GetNext(CurItem), GetPrev(CurItem));
	delete GetItem(CurItem);
	delete CurItem;
	CurItem = CurItemTmp;
      }          
    }
    CurItem=GetNext(CurItem);
  }
  return;
}






template <class ITEM>
ITEM *SORTLIST<ITEM>::Find(ITEM *Item)
{
LISTITEM<ITEM> *CurItem;
ITEM *RetItem;
if (Item == NULL) return NULL;

// Find where the item is
CurItem=Front;
while (   (CurItem != NULL)  && (*GetItem(CurItem) != *Item))
    {
    CurItem=GetNext(CurItem);
    }
if (CurItem!=NULL && *GetItem(CurItem) == *Item)
    RetItem=GetItem(CurItem);
else
    RetItem=NULL;
return(RetItem);
}


template<class ITEM>
void SORTLIST<ITEM>::Print()
{
cout << "\t   List = ";
LISTITEM<ITEM>* h = Front;
while (h != NULL)
  { h->Item->Print();
    cout << " ";
    h = h->Next;
  }
cout << endl;
}


#endif
