// *******************************************************************
//
//  A Teaching Environement for Remote Lecturing in Higher Education
//
//         Prototype Implementation of a  Scalable Multicast Protocol
//
//                          Werner Geyer
//                      University of Mannheim
//                     Praktische Informatik IV
//                              L 15, 16
//	                    68131 Mannheim
//                             Germany
//
//              geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
//
//  Copyright (c) 1996, 1997, 1998
//
//     Lehrstuhl Praktische Informatik IV
//     University of Mannheim
//     Germany
//
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this
//  software and its documentation for any purpose and without
//  fee is hereby granted, provided that the above copyright
//  notice appear in all copies and that both that copyright
//  notice and this permission notice appear in supporting
//  documentation.
//
//  We make no representations about the suitability of this
//  software for any purpose.  It is provided "as is" without
//  express or implied warranty.
//
//
//  Author: Werner Geyer
//          Praktische Informatik IV
//          University of Mannheim
//          email: geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
// program : SMP.cpp
// authors : Martin Grumann, Juergen Vogel, Werner Geyer
// date    : Aug 1996
// update  : Feb 1997, Jan 1998
// desc    : SMP library for linking to application
// *******************************************************************
#include "general.h"

#include "SMP.h"
#include "iohelpers.h"
#include "Demon.h"


#ifdef WIN32

#define xHANDLE_SIGKILL

extern "C" {
  void clearReadSocketMessageQueue(int fd);
  void restoreReadSocketMessageQueue(int fd);
}
#endif


extern "C" {
  void linksocket(int fd, int mask, Tcl_FileProc * callback);
  void unlinksocket(int fd);
}


SMP::SMP* SMP::SMP::smp_pointer;

//FILE* SMP::SMP::debug = NULL;
FILE* SMP::SMP::debug = stdout;


/*********************************************************************
 Name    : SMP::SMP
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
SMP::SMP::SMP() : groups(0), ErrNo(0) {
#ifdef WIN32
  // initialize random generator
  srand( (unsigned)time( NULL ) );
#endif

  for (int x = 0; x < MAX_CLIENTS; x++) {
    belegt[x] = false;
    loggedOn[x] = false;
  }

  smp_pointer = this;


  // SIGSEGV abfangen
#ifdef HANDLE_SIGKILL
  signal(SIGSEGV, ExitHandler);
#ifdef WIN32
#else
  signal(SIGKILL, ExitHandler);
#endif
#endif

  //debug = fopen("debugsmp.txt","w");

  // get local adresses
  struct hostent *host = NULL;
  struct in_addr a;
  const int len = 200;
  char *c = (char *) malloc(len);
  gethostname(c, len);
  host = gethostbyname(c);
  free(c);
  if ((host != NULL) && (host->h_addr_list[0])) {
    memcpy(&(a.s_addr), host->h_addr_list[0], host->h_length);
    ipAddr = ntohl(a.s_addr);
  } else {
    host = NULL;
    host = gethostbyname("localhost");
    if ((host != NULL) && (host->h_addr_list[0])) {
      memcpy(&(a.s_addr), host->h_addr_list[0], host->h_length);
      ipAddr = ntohl(a.s_addr);
    } else {
      std::cerr << "Can not resolve local address " << std::endl;
      ipAddr = 0;
    }
  }

  Demon::Deamon = new Demon();
}

#ifdef WIN32
bool SMP::SMP::StartupSockets() {
  WSADATA WSAdata;
  /* XXX
   * initialize our socket interface plus the tcl 7.5 socket
   * interface (since they redefine some routines we call).
   * eventually we should just call the tcl sockets but at
   * the moment that's hard to set up since they only support
   * tcp in the notifier.
   */
  if (WSAStartup(MAKEWORD(2, 0), &WSAdata)) {
    perror("SMP::SMP - Windows Sockets init failed");
    std::cerr << WSAGetLastError() << std::endl;
    return false;
  }

  /*
   * Increase the application queue size from default value of 8.
   * At the default value, cross application SendMessage of WM_KILLFOCUS
   * will fail because the handler will not be able to do a PostMessage!
   * This is only needed for Windows 3.x, since NT dynamically expands
   * the queue.
   */
  SetMessageQueue(64);

  return true;
}
#endif


/*********************************************************************
 Name    : SMP::~SMP
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
SMP::SMP::~SMP() {
  for (int x = 0; x < MAX_CLIENTS; x++)
    if (( belegt[x] == true) && (loggedOn[x] == true)) {
      LeaveGroup(x);
      Close(x);
    }
  //fclose(debug);
  Demon::Deamon->ShutDown(1);
}

/*********************************************************************
 Name    : SMP::allocate
 Param   :
 Result  : index of free docking station for application
           -1 if max. number of group memberships is reached
 Desc    :
 Calledby: SMP::Open
 Uses    :
 *********************************************************************/
int SMP::SMP::allocate() {
  if (groups >= MAX_CLIENTS) return -1;
  int index = -1;
  while (belegt[++index]);
  return index;
}


/*********************************************************************
 Name    : SMP::Open
 Param   : port DaemonPort for incoming calls
 Result  : socket-descriptor if successful, otherwise -1
 Desc    : connects SMP-Library with SMP-Deamon
 Calledby: Application
 Uses    :
 *********************************************************************/
int SMP::SMP::Open(void (*Callback)(Indication), void* param) {

#ifndef WIN32
  int sflags;
#endif

  int toHandle = allocate();
  if (toHandle < 0) {
    std::cerr << "SMP::Open() : group overflow - could not connect to smp" << std::endl;
    ErrNo = GROUP_OVERFLOW;
    return -1;
  }

  //std::cout << "toHandle = " << toHandle << std::endl;

  groups++;
  belegt[toHandle] = true;

  group[toHandle].Callback = Callback;
  group[toHandle].Param    = param;

  // init demon
  int newPos = Demon::Deamon->c.search_pos();
  Demon::Deamon->c.accepted[newPos] = true;

  return toHandle;
}


/*********************************************************************
 Name    : SMP::JoinGroup
 Param   : sock      socket-descriptor delivered by SMP::open as indicator
                     for wanted group
           mcAddr    multicast group address in point notation
           mcPort    port number for SMP-Deamon to listen
           TTL       TimeToLive for sended packets
           Bwidth    netrate in bytes per second
           SClass    Service Classes A,B,C
           SOrdering if true, Source Ordering is enabled
           LJoin     if true, Late Join is enabled
           SData     if true, application and SMP use Shared Memory
 Result  : true, if successful, otherwise false
 Desc    : let application join group
           before used, application must call SMP::Open
 Calledby: application
 Uses    :
 *********************************************************************/
bool SMP::SMP::JoinGroup(int toHandle,
                    const char* mcAddr,
                    int mcPort,
                    int TTL,
                    int Bwidth,
                    char SClass,
                    bool SOrdering,
                    bool LJoin,
                    bool SData,
		    bool loopback) {


  //std::cout << "SMP::JoinGroup" << std::endl;

  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::JoinGroup() : BAD_SOCKET_DESC" << std::endl;
    return false;
  }

  group[toHandle].LateJoin = LJoin;
  group[toHandle].SharedData = false;   // = SData;
  group[toHandle].loopback = loopback;

  LogOnInfo* info = new LogOnInfo();
  strcpy(info->multicastAddr, mcAddr);
  info->multicastPort = mcPort;
  info->ttl = TTL;
  info->ServiceClass = SClass;
  info->SourceOrdering = SOrdering;
  info->LateJoin = LJoin;
  info->SharedData = SData;
  info->Bandwidth = Bwidth;
  info->size = 0;
  info->Loopback = loopback;

  loggedOn[toHandle] = true;

  Demon::Deamon->logOn(toHandle, info);

  return true;
}

/*********************************************************************
 Name    : SMP::Send
 Param   : sock     socket-descriptor delivered by SMP::open as indicator
                    for wanted group
           ADU      memory addess of ADU wanted to send
           ADUSize  size of ADU (without Header)
 Result  : true, if successful, otherwise false
 Desc    : send ADU to group
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::Send(int toHandle, void* ADU, int ADUSize) {

#ifdef _SMP_DATATRACE
    std::cout << "SMP::Send(" << toHandle << ") : application sends unreliable data"
							<< std::endl;
#endif

  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::Send() : BAD_SOCKET_DESC" << std::endl;
    return false;
  }

  int size   = ADUSize + sizeof(SMPHeader);
  char* data = Demon::Deamon->AllocatePacketBuffer(toHandle, size);

  memcpy(data + sizeof(SMPHeader), (char*)ADU, ADUSize);

  if (Demon::Deamon->SendData(toHandle, data, size) < 0) {
    std::cerr << "SMP::Send(" << toHandle << ") : could not send data"
	 << std::endl;
    Demon::Deamon->FreePacketBuffer(toHandle, data, size, false);
    return false;
  }

  if (group[toHandle].loopback) {
    Indication ind;
    ind.ServiceType = Indication::DATAIND;
    ind.buf = new Buffer(ADUSize);
    memcpy(ind.buf->getCont(), (char*)ADU, ADUSize);
    ind.param = group[toHandle].Param;
    ind.senderIPAddr = ipAddr;
    group[toHandle].Callback(ind);
  }

  return true;
}



/*********************************************************************
 Name    : SMP::SendUnreliable
 Param   : sock     socket-descriptor delivered by SMP::open as indicator
                    for wanted group
           ADU      memory addess of ADU wanted to send
           ADUSize  size of ADU (without Header)
 Result  : true, if successful, otherwise false
 Desc    : send ADU UNRELIABLE to group
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::SendUnreliable(int toHandle, void* ADU, int ADUSize) {

#ifdef _SMP_DATATRACE
    std::cout << "SMP::SendUnreliable(" << toHandle << ") : application sends unreliable data"
	 << std::endl;
#endif

  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::Send() : BAD_SOCKET_DESC" << std::endl;
    return false;
  }

  int size   = ADUSize + sizeof(SMPHeader);
  char* data = Demon::Deamon->AllocatePacketBuffer(toHandle, size);

  memcpy(data + sizeof(SMPHeader), (char*)ADU, ADUSize);

  if (Demon::Deamon->SendUnreliableData(toHandle, data, size) < 0) {
    std::cerr << "SMP::Send(" << toHandle << ") : could not send data"
	 << std::endl;
    Demon::Deamon->FreePacketBuffer(toHandle, data, size, false);
    return false;
  }

  if (group[toHandle].loopback) {
    Indication ind;
    ind.ServiceType = Indication::DATAIND;
    ind.buf = new Buffer(ADUSize);
    memcpy(ind.buf->getCont(), (char*)ADU, ADUSize);
    ind.param = group[toHandle].Param;
    ind.senderIPAddr = ipAddr;
    group[toHandle].Callback(ind);
  }

  return(true);
}


/*********************************************************************
 Name    : SMP::SendDelayedADURequest
 Param   : sock     socket-descriptor delivered by SMP::open as indicator
                    for wanted group
           ADU      memory addess of ADU wanted to send
           ADUSize  size of ADU (without Header)
 Result  : true, if successful, otherwise false
 Desc    : send DelayedADU-Request to group
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::SendDelayedADURequest(int toHandle, void* ADU, int ADUSize) {
  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::SendDelayedADURequest() : BAD_SOCKET_DESC" << std::endl;
    return false;
  }

  if (!Demon::Deamon->c.LateJoin[toHandle]) {

#ifdef _SMP_LJ
    std::cout << "Demon::SendDelayedADURequest(" << toHandle << ") : application "
	 << "requests delayed adu" << std::endl;
#endif

    int size   = ADUSize + sizeof(SMPHeader);
    char* data = Demon::Deamon->AllocatePacketBuffer(toHandle, size);

    memcpy(data + sizeof(SMPHeader), (char*)ADU, ADUSize);

    //	std::cout << "DADU request: arrived = " << com.size << " size = "
    //	     << size << std::endl;

    if (Demon::Deamon->c.OpenRequest[toHandle]) {
#ifdef _SMP_LJ
      std::cout << "SMP::SendDelayedADURequest(" << toHandle << ") : open request ..." << std::endl;
#endif
      Demon::Deamon->FreePacketBuffer(toHandle, data, size, false);
      return false;
    }

    // *******************************************************************
    // store dadu req in buffer
    // *******************************************************************
    Demon::Deamon->c.DAduRequest[toHandle].ssrc = Demon::Deamon->c.We[toHandle]->SSRC;
    Demon::Deamon->c.DAduRequest[toHandle].sn = ++(Demon::Deamon->c.dadu_seqno[toHandle]);
    Demon::Deamon->c.DAduRequest[toHandle].data = (char*)data;
    Demon::Deamon->c.DAduRequest[toHandle].size = size;

    if (Demon::Deamon->SendDADUReq(toHandle, data, size) < 0) {

#ifdef _SMP_LJ
      std::cerr << "SMP::SendDelayedADURequest(" << toHandle << ") : could not send DADUReq"
	   << std::endl;
#endif

    }
    Demon::Deamon->SetTimer(toHandle, Demon::Deamon->c.We[toHandle]->SSRC,
		    0, Timer::REPEAT, Demon::DEFAULTDISTANCE,
		    Demon::Deamon->c.We[toHandle], Demon::Deamon->c.We[toHandle]->SSRC,
		    0, true, 0, false, 0);

    Demon::Deamon->c.OpenRequest[toHandle] = true;
  }

  return(true);
}

/*********************************************************************
 Name    : SMP::SendDelayedADUResponse
 Param   : sock     socket-descriptor delivered by SMP::open as indicator
                    for wanted group
           ADU      memory addess of ADU wanted to send
           ADUSize  size of ADU (without Header)
 Result  : true, if successful, otherwise false
 Desc    : send DelayedADU-Response to group
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::SendDelayedADUResponse(int toHandle, void* ADU, int ADUSize) {
  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::SendDelayedADUResponse() : BAD_SOCKET_DESC" << std::endl;
    return false;

  }

  if (!Demon::Deamon->c.LateJoin[toHandle]) {

#ifdef _SMP_LJ
    std::cout << "SendDelayedADUResponse(" << toHandle << ") : application "
	 << "responds to delayed adu request" << std::endl;
#endif

    int size   = ADUSize + sizeof(SMPHeader);
    char* data = Demon::Deamon->AllocatePacketBuffer(toHandle, size);

    memcpy(data + sizeof(SMPHeader), (char*)ADU, ADUSize);

    //	std::cout << "DADU response: arrived = " << com.size << " size = "
    //	     << size << std::endl;

    if (Demon::Deamon->SendDADURsp(toHandle, data, size, 0) < 0) {
      std::cerr << "SendDelayedADUResponse(" << toHandle << ") : could not send DADURsp"
	   << std::endl;
      return false;
    }
  }

  return true;
}

/*********************************************************************
 Name    : SMP::DelayedADUResponseNotPossible
 Param   : sock     socket-descriptor delivered by SMP::open as indicator
 Result  : true, if successful, otherwise false
 Desc    : the received dadu request cannot be answerd
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::DelayedADUResponseNotPossible(int toHandle, unsigned int id) {
  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
      std::cerr << "SMP::DelayedADUResponseNotPossible() : BAD_SOCKET_DESC" << std::endl;
      return false;
  }

#ifdef _SMP_LJ
  std::cout << "SMP::DelayedADUResponseNotPossible(" << toHandle
       << ") : application cannot answer dadu request - clear buffer" << std::endl;
#endif

  DADU * dadu = Demon::Deamon->c.FindDADU(toHandle, id);
  if (dadu) {
    DADU toDelete(dadu->ssrc, dadu->sn);
    delete(Demon::Deamon->c.dadulist[toHandle].Delete(&toDelete));
  } else {

#ifdef _SMP_LJ
    std::cerr << "SMP::DelayedADUResponseNotPossible() : could not find DADU request "
	 << "(already discarded)" << std::endl;
#endif
  }

  return true;
}

/*********************************************************************
 Name    : SMP::DelayedADUResponsePossible
 Param   : sock     socket-descriptor delivered by SMP::open as indicator
 Result  : true, if successful, otherwise false
 Desc    : the received dadu request cannot be answerd
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::DelayedADUResponsePossible(int toHandle, unsigned int id) {
  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
      std::cerr << "SMP::DelayedADUResponseNotPossible() : BAD_SOCKET_DESC" << std::endl;
      return false;
  }

#ifdef _SMP_LJ
  std::cout << "SMP::DelayedADUResponsePossible(" << toHandle
       << ") : application can answer dadu request - clear buffer" << std::endl;
#endif

  int size = sizeof(SMPHeader) + sizeof(SMPDelADURespBody);
  char * data = Demon::Deamon->AllocatePacketBuffer(toHandle, size);
  if (Demon::Deamon->SendDADURsp(toHandle, data, size, id) < 0) {
    std::cerr << "SMP::DelayedADUResponsePossible(" << toHandle << ") : could not send DADURsp"
	 << std::endl;
    return false;
  }


  return(true);
}

/*********************************************************************
 Name    : SMP::LeaveGroup
 Param   : sock   socket-descriptor delivered by SMP::open as indicator
                  for wanted group
 Result  : true, if successful, otherwise false
 Desc    : application leaves group
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::LeaveGroup(int toHandle) {
  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::LeaveGroup() : BAD_SOCKET_DESC" << std::endl;
    return false;
  }

#ifdef _SMP_DATATRACE
  std::cout << "SMP::LeaveGroup(" << toHandle << ") : received logoff" << std::endl;
#endif

  Demon::Deamon->logOff(toHandle);

  loggedOn[toHandle] = false;

  return true;
}

/*********************************************************************
 Name    : SMP::Close
 Param   : sock   socket-descriptor delivered by SMP::open as indicator
                  for wanted group
 Result  : true, if successful, otherwise false
 Desc    : close connection between SMP-Library and SMP-Daemon for this
           group
 Calledby: Application
 Uses    :
 *********************************************************************/
bool SMP::SMP::Close(int toHandle) {
  if ((toHandle < 0) || (toHandle >= MAX_CLIENTS)) {
    std::cerr << "SMP::Close() : BAD_SOCKET_DESC" << std::endl;
    return false;
  }

  if (loggedOn[toHandle]) LeaveGroup(toHandle);

#ifdef _SMP_DATATRACE
  std::cout << "SMP::Close(" << toHandle << ") : received close" << std::endl;
#endif

  Demon::Deamon->close(toHandle);

  belegt[toHandle] = false;
  groups--;
  return true;
}



/*********************************************************************
 Name    : SegmentationFaultHandler
 Param   :
 Result  :
 Desc    : is called if a segmentation fault occured - informs Daemon
           by logging off all clients, then shuts down application

 Calledby: os
 Uses    :
 *********************************************************************/
void SMP::ExitHandler(int sig) {
  for (int i = 0; i < MAX_CLIENTS; i++) {
    if (!SMP::smp_pointer) return;
    if (SMP::smp_pointer->loggedOn[i]) {
      SMP::smp_pointer->LeaveGroup(i);
      SMP::smp_pointer->Close(i);
    }
  }

  Demon::Deamon->ShutDown(1);

  std::cerr << "ExitHandler() : fatal Error, smp has been informed" << std::endl;
}




