// *******************************************************************
//
//  A Teaching Environement for Remote Lecturing in Higher Education
//
//         Prototype Implementation of a  Scalable Multicast Protocol
//
//                          Werner Geyer
//                      University of Mannheim
//                     Praktische Informatik IV
//                              L 15, 16
//	                    68131 Mannheim
//                             Germany
//
//              geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
//
//  Copyright (c) 1996, 1997
//
//     Lehrstuhl Praktische Informatik IV
//     University of Mannheim
//     Germany
//
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this
//  software and its documentation for any purpose and without
//  fee is hereby granted, provided that the above copyright
//  notice appear in all copies and that both that copyright
//  notice and this permission notice appear in supporting
//  documentation.
//
//  We make no representations about the suitability of this
//  software for any purpose.  It is provided "as is" without
//  express or implied warranty.
//
//
//  Author: Werner Geyer
//          Praktische Informatik IV
//          University of Mannheim
//          email: geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
// program : SMP.h
// authors : Martin Grumann
// date    : Aug 1996
// update  : Feb 1997
// desc    : SMP library for linking to application
// *******************************************************************
#ifndef __SMP_H
#define __SMP_H


extern "C" {

#include <signal.h>   // signal
#include <stdio.h>    // perror
#include <errno.h>    // perror
#include <string.h>  // strncmp
#include <stdlib.h>   // atoi

#ifdef WIN32
#include <time.h>
#include <io.h>
#else
#include <arpa/inet.h>
#include <sys/un.h>
#include <fcntl.h>
#endif
}

#include <tcl.h>

#include <util/Buffer.h>
#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include <iostream> // cout

#include "common.h"
#include "SMPHeader.h"

namespace SMP {

void ExitHandler(int sig);

class Indication {

public:
  Indication() : ServiceType(0), Id(0), senderIPAddr(0), buf(NULL) {}

  int ServiceType;
  unsigned int Id;
  unsigned int senderIPAddr;
  RefCount<Buffer> buf;
  void* param;

  enum {
    NOSERVICE = 0,
    JOININD   = 1,
    LEAVIND   = 2,
    DATAIND   = 3,
    UNRLDATAIND = 13,
    ACKIND    = 4,
    ERR       = 5,
    LJNEEDED  = 7,     // lj: we are too late: Late Join needed
    DADUCONF  = 8,     // lj: Confirmation for DelayedADU
    DADUIND   = 9,     // lj: Indication for DelayedADU
    CLOSEIND  = 10,
    ALIVE     = 11,
    DADUFAIL  = 12     // lj: Our DelayedADURequest failed
  };
};


typedef void (*SMPCallback) (Indication);


class SMP {

public:
  SMP();
  ~SMP();

  // smp service interface
  int  Open(void (*callback)(Indication), void* param= NULL);
  bool Close(int socket);
  bool JoinGroup(int socket, const char* mcAddress, int port, int ttl, int bandwidth, char serviceClass,
		 bool sourceOrdering, bool lateJoin, bool sharedData, bool loopback = false);
  bool LeaveGroup(int socket);
  bool Send(int socket, void* dataAddr, int dataSize);
  bool SendUnreliable(int socket, void* dataAddr, int dataSize);
  bool SendDelayedADURequest(int socket, void* dataAddr, int dataSize);
  bool SendDelayedADUResponse(int socket, void* dataAddr, int dataSize);
  bool DelayedADUResponseNotPossible(int socket, unsigned int id);
  bool DelayedADUResponsePossible(int socket, unsigned int id);

  // internal stuff
#ifdef WIN32
  static bool StartupSockets();
#endif

  void Error(char* = NULL);
  int ErrNo;
  enum { BAD_PATH = 1,
         BAD_PROJECT = 2,
         GROUP_OVERFLOW = 3,
         BAD_SOCKET_DESC = 4,
         TOO_MANY_CLIENTS = 5,
         BAD_OPERATION_MODE = 6,
         BROKEN_CONNECTION = 7,
         NEW_NOT_AV = 8,
         DEL_NOT_AV = 9,
         SYS_ERR = 100
  };

  static SMP* smp_pointer;

  bool loggedOn[MAX_CLIENTS];

  struct {
    SMPCallback Callback;
    void* Param;
    bool LateJoin;
    bool SharedData;
    bool loopback;
    char path_name[64];
  } group[MAX_CLIENTS];

private:
  int allocate();

  int groups;
  bool belegt[MAX_CLIENTS];
  char hostInAddr[16];
  unsigned int ipAddr;

 public:
  static FILE* debug;
};
}

#endif

