// *******************************************************************
//
//  A Teaching Environement for Remote Lecturing in Higher Education
//
//         Prototype Implementation of a  Scalable Multicast Protocol
//
//                          Werner Geyer
//                      University of Mannheim
//                     Praktische Informatik IV
//                              L 15, 16
//	                    68131 Mannheim
//                             Germany
//
//              geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
//
//  Copyright (c) 1996, 1997
//
//     Lehrstuhl Praktische Informatik IV
//     University of Mannheim
//     Germany
//
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this
//  software and its documentation for any purpose and without
//  fee is hereby granted, provided that the above copyright
//  notice appear in all copies and that both that copyright
//  notice and this permission notice appear in supporting
//  documentation.
//
//  We make no representations about the suitability of this
//  software for any purpose.  It is provided "as is" without
//  express or implied warranty.
//
//
//  Author: Werner Geyer
//          Praktische Informatik IV
//          University of Mannheim
//          email: geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
// program : SMPHeader.h
// authors : Juergen Vogel
// date    : Jul 1997
// update  :
// desc    : Defines Header and Bodies for SMP
// *******************************************************************
#ifndef _SMPHEADER_H
#define _SMPHEADER_H


#include <iostream> // cout
#include "mytypes.h"

//#define DATA     0
//#define DADURQ 100
//#define DADURP 101
//#define JOIN   124
//#define LEAV   125
//#define SMES   126
//#define REQ    127

#define XADDR 101
#define MARKEROFF 0
#define MARKERON 1

#define MAX_MEMBERS_PER_SESSION_MESSAGE 15

#define SMP_PT 0x7FFF
#define SMP_X  0x8000

/* Payload Types :
     0 Anwendungsdaten
   100 DelayedADURequest
   101 DelayedADUResponse
   124 Join-Message
   125 Leave-Message
   126 Session-Message
   127 NACK
*/

namespace SMP {
class SMPHeader {
 public:
  enum type {
    DATA=0,
		DATA_UNRELIABLE = 1,
    DADURQ=100,
    DADURP=101,
    JOIN=124,
    LEAV=125,
    SMES=126,
    REQ=127
  };
	
  u_int16 FLAGS; // X:1 PT:15 (PT = payload type */
  u_int16 SZ;    // Packet size
  u_int32 TS;    // Timestamp
  u_int32 SSRC;  // Identifier
  u_int32 ORGIP; // original senders's ip
  u_int32 SN;    // Sequence Number
};

class SMPDelADURespBody {
public:
  u_int32 RSSRC;   // SSRC of intentioned receiver
};

class SMPSessionMessageCount {
public:
  u_int32 i;
};

class SMPSessionMessageElement {
public:
  u_int32 SSRC;
  u_int32 Timestamp1;
  u_int32 Delta;
  u_int32 SN;
  u_int32 HSN;
  u_int16 Active;
};

class SMPRequestBody {
public:
  u_int32 ReqSSRC;
  u_int32 ReqSN;
  u_int16 ReqTTL;
};
}

#endif



