// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// Copyright (c) 1997 University of Mannheim. All Rights Reserved.
//
// Permission to use, copy, modify, and distribute this software and its 
// documentation for educational, research and non-profit purposes, 
// without fee, and without a written agreement is hereby granted, 
// provided that the above copyright notice, this paragraph and the 
// following four paragraphs appear in all copies.  
// 
// Permission to incorporate this software into commercial products may 
// be obtained from the following address:
// University of Mannheim
// Lehrstuhl fuer Praktische Informatik IV
// L15, 16
// 68161 Mannheim, Germany
// 
// IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY 
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, 
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND 
// ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED 
// OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
// AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER 
// IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS 
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
// 
// The University of Mannheim will appreciate receiving any enhancements and 
// modifications of this software made by others, and reserves the right to 
// include them in future releases. The University also recommends that this 
// software not be redistributed by third parties, but be obtained by all 
// directly from the University's server, so that all users of the software 
// may receive bug notifications and other relevant announcements.
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// $Id: TclTimer.h,v 1.0 1997/04/17 17:48:27 whd Exp $
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#ifndef _TCL_TIMER_H_
#define _TCL_TIMER_H_

extern "C" {
#include <sys/time.h>
}

#include <tcl.h>

#include <String.h>
#include <DLList.h>

namespace SMP {
class TclTimerInterface {
 public:
  virtual void timerHandler(String) =0;
};
typedef TclTimerInterface *P2TclTimerInterface;

class TclTimerToken {
 public:
  Tcl_TimerToken token;
  ClientData	 cd;
  Pix		 index;
};
typedef TclTimerToken *P2TclTimerToken;

class TclTimer {

public:
  TclTimer(String ="noname", P2TclTimerInterface =NULL);
  ~TclTimer();

  String getId() { return id; };
  void setId(String i) { id=i; };

  P2TclTimerInterface getPif() { return pif; };
  void setPif(P2TclTimerInterface p) { pif=p; };

  inline void msched(int msecs) { addToken(msecs); };
  inline void usched(int usecs) { addToken(usecs/1000); };
  inline void ssched(int secs)  { addToken(secs*1000); };

  inline int isActive() { return(token_list.length()>0); };
  void cancel();

private:
  P2TclTimerInterface	pif;
  String		id;
  DLList<TclTimerToken>	token_list;

  void addToken(int);
  static void dispatch(ClientData);

};
typedef TclTimer *P2TclTimer;
}

#endif







