// *******************************************************************
//
//  A Teaching Environement for Remote Lecturing in Higher Education
//
//         Prototype Implementation of a  Scalable Multicast Protocol
//
//                          Werner Geyer
//                      University of Mannheim
//                     Praktische Informatik IV
//                              L 15, 16
//	                    68131 Mannheim
//                             Germany
//
//              geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
//
//  Copyright (c) 1996, 1997
//
//     Lehrstuhl Praktische Informatik IV
//     University of Mannheim
//     Germany
//
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this
//  software and its documentation for any purpose and without
//  fee is hereby granted, provided that the above copyright
//  notice appear in all copies and that both that copyright
//  notice and this permission notice appear in supporting
//  documentation.
//
//  We make no representations about the suitability of this
//  software for any purpose.  It is provided "as is" without
//  express or implied warranty.
//
//
//  Author: Werner Geyer
//          Praktische Informatik IV
//          University of Mannheim
//          email: geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
// program : Timer.cpp
// authors : Martin Grumann
// date    : Aug 1996
// update  : Feb 1997
// desc    : implementation of a Timer
// *******************************************************************

#ifndef __TIMER_CPP
#define __TIMER_CPP


#include <iostream> // cout
#include <stdio.h>
#include "SMP.h"
#include "Timer.h"


unsigned int GetTimestamp() {
#ifdef WIN32
   struct _timeb timebuffer;
   _ftime(&timebuffer);
	 return (1000*timebuffer.time + timebuffer.millitm);
#else
  struct timeval tp;
  gettimeofday(&tp, 0);
  return ((tp.tv_sec * 1000) + (tp.tv_usec / 1000));
#endif
}


/*********************************************************************
 Name    : rtcp_intervall
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
double rtcp_intervall(int members,
                      double rtcp_bw,
                      int packet_size,
                      double* avg_rtcp_size,
                      bool initial) {

  double const RTCP_MIN_TIME = 5.0;        // Seconds
  double const RTCP_SIZE_GAIN = (1.0/16.0);
  double ret;
  double rtcp_min_time = RTCP_MIN_TIME;

  if (initial) {
      rtcp_min_time /= 2;
      *avg_rtcp_size = 76;
     // 28 (UDP/IP) + 12(RTP Header) + 4 (Length) + 2 (#Members)*16(SM Element)
  }

  *avg_rtcp_size += (packet_size - *avg_rtcp_size)*RTCP_SIZE_GAIN;

  ret = (*avg_rtcp_size)*members / rtcp_bw;
  if (ret < rtcp_min_time)
     ret = rtcp_min_time;

  return ret * (drand48() + 0.5);
}


/*********************************************************************
 Name    : Min
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
int Min(int i1, int i2) {
  return  (i1 < i2) ?  i1 :  i2;
}



/*********************************************************************
 Name    : Timer::Timer
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
SMP::Timer::Timer(int toH, unsigned int ssrc,
             unsigned int seq, int action, int delay, Member* m,
             unsigned int Erz, double Laufzeit, bool Locally, signed
             char ttl, bool FNack, int count)
  : C1(1), C2(2), D1(1), D2(2),

  //C1(0.9), C2(1.4), D1(0.5), D2(1.0),
    toHandle(toH), SSRC(ssrc),
    SN(seq), ActionId(action), Delay(delay),
    Sender(m), Erzeuger(Erz),
    LocalRecovery(Locally), TTL(ttl), ForeignNack(FNack), i(count){
  
  // StartZeitPunkt initialisieren
  StartZeitPunkt = GetTimestamp();
  
  // GesamtLaufZeit errechnen
  switch (ActionId) {
  case REPAIR:
    GesamtLaufZeit = GetRepairTimer(Delay, LocalRecovery);
    break;
    
  case REQUEST:
    GesamtLaufZeit = GetRequestTimer(Delay, i);
    break;
    
  case SESSMESS:
    // Die Lauzeit des Timers enthaelt der Parameter Laufzeit als double
    GesamtLaufZeit = (unsigned int)(Laufzeit * 1000);
    break;
    
  case RATECTL:
    GesamtLaufZeit = 100;
    break;
    
  case REPEAT: // lj
    GesamtLaufZeit = GetRequestTimer(Delay, i);

#ifdef _SMP_LJ
    cout << "Laufzeit REPEAT = " << GesamtLaufZeit << " ms" << endl;
#endif
    break;
    
  case GIVETHROUGH: // lj

    if (i>=1) {
      GesamtLaufZeit = 1000 * 10;
    } else {
      GesamtLaufZeit = GetRepairTimer(Delay, LocalRecovery);
    }

#ifdef _SMP_LJ
    cout << "Laufzeit GIVETROUGH = " << GesamtLaufZeit << " ms" << endl;
#endif
    break;
    
  case DELAYEDSEND:
    GesamtLaufZeit = 150;
    break;

  case DELAYEDSENDU:
    GesamtLaufZeit = 150;
    break;

  default:
#ifdef _SMP_DEBUG
    cout << "new timer: wrong action id " << ActionId << endl;
#endif
    break;
  }

  // EndZeitPunkt errechnen
  EndZeitPunkt = StartZeitPunkt + GesamtLaufZeit;
}


/*********************************************************************
 Name    : Timer::ReUse
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
void SMP::Timer::ReUse(int toH, unsigned int ssrc,
             int seq, int action, int delay, Member* sender,
             unsigned int Erz, double Laufzeit, bool Locally,
             signed char ttl, bool FNack, int count) {


  toHandle = toH;
  SSRC = ssrc;
  SN = seq;
  ActionId = action;
  Delay = delay;
  Sender = sender;
  Erzeuger = Erz;
  LocalRecovery = Locally;
  TTL = ttl;
  ForeignNack = FNack;
  i = count;

  // C1, C2, D1, D2 bleiben so wie im Konstruktor gesetzt

    // StartZeitPunkt initialisieren
    StartZeitPunkt = GetTimestamp();

    // GesamtLaufZeit errechnen
    switch (ActionId) {
    case REPAIR:
         GesamtLaufZeit = GetRepairTimer(Delay, LocalRecovery);
         break;

    case REQUEST:
         GesamtLaufZeit = GetRequestTimer(Delay, i);
         break;

    case SESSMESS:
         // Die Lauzeit des Timers enthaelt der Parameter Laufzeit als double
         GesamtLaufZeit = (unsigned int)(Laufzeit * 1000);
         break;

    case RATECTL:
         GesamtLaufZeit = 100;
         break;

    case REPEAT: // lj
         GesamtLaufZeit = GetRequestTimer(Delay, i);

#ifdef _SMP_LJ
	 cout << "Laufzeit REPEAT = " << GesamtLaufZeit << " ms" << endl;
#endif
          break;

    case GIVETHROUGH: // lj

      if (i>=1) {
	GesamtLaufZeit = 1000 * 10;
      } else {
	GesamtLaufZeit = GetRepairTimer(Delay, LocalRecovery);
      }

#ifdef _SMP_LJ
	 cout << "Laufzeit GIVETROUGH = " << GesamtLaufZeit << " ms" << endl;
#endif
         break;

    default:
      //cerr << "reuse timer: wrong action id " << ActionId << endl;
      break;
   }

   // EndZeitPunkt errechnen
   EndZeitPunkt = StartZeitPunkt + GesamtLaufZeit;
}


/*********************************************************************
 Name    : Timer::operator==
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
bool SMP::Timer::operator==(const Timer& t){
  if ((Erzeuger == t.Erzeuger) &&
      (SSRC == t.SSRC) &&
      (SN == t.SN) &&
      (ActionId == t.ActionId))
    return true;
  return false;
}


/*********************************************************************
 Name    : Timer::IsEqual
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
bool SMP::Timer::IsEqual(const Timer& t) {

  // Wenn Erzeuger angegeben ist, dann will eine Applikatio
  // den deamon verlassen, wenn er gleich null ist, verlaesst ein
  // (fremdes) Mitglied die Gruppe



  if (t.Erzeuger == 0) {
    if (SSRC == t.SSRC) {

#ifdef _SMP_TIMER
      cout << "Timer::IsEqual() : isequal (nicht Erzeuger): type = "
	   << ActionId << endl;
#endif

      return true;
    }
    else return false;
  }
  else {
    if (Erzeuger == t.Erzeuger) {

#ifdef _SMP_TIMER
      cout << "Timer::IsEqual() : isequal (Erzeuger): type = "
	   << ActionId << endl;
#endif

      return true;
    }
    else return false;
  }
}


/*********************************************************************
 Name    : Timer::operator!=
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
bool SMP::Timer::operator!=(const Timer& t){
  return !operator==(t);
}


/*********************************************************************
 Name    : Timer::operator
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
bool SMP::Timer::operator<(const Timer& t){
  return(EndZeitPunkt < t.EndZeitPunkt);
}


/*********************************************************************
 Name    : Timer::GetRequestTimer
 Param   :
 Result  :
 Desc    : Erechnet die Laufzeit von Request Timer
           Parameter enthaelt die Distance zum Sender in Millisekunden
           In i ist spezifiziert, wie oft bereits ein Timer (Request)
           fuer ein bestimmtes Paket gestellt wurde.
 Calledby:
 Uses    :
 *********************************************************************/
unsigned int SMP::Timer::GetRequestTimer(unsigned int Distance, int i) {

  unsigned int ret;
  double  d =  drand48();
  double dist = Distance/1000.; // Dist enthaelt Distanz in Sekunden

#ifdef _SMP_REQTIMER
  cout << "i = " << i << " Distance = " << Distance << " drand48() = " << d
       << " SN = " << SN << " toHandle = " << toHandle;
#endif

  // was i == 0, changed to i == -1 in oder to avoid local request
  // sense?
  if (i == -1) {
     // Lokaler Request
     // Laufzeit ist nur vom Zufall abhaengig
    //ret.tv_sec = 0;
    //d *= 1000000; // d enthaelt jetzt die Anzahl der Mikrosekunden
    //ret.tv_usec = (int)d;
  }
  else {

     // Globales Nack mit i.ter Wiederholung
     // Laufzeitbestimmung wie in SRM

     if (dist <= 0)
         dist = 0.1;

     // d E [C1*Dist; (C1 +C2)*Dist] C1*dist
     // d =  (C1 * dist) + d * ((C1 + C2) * dist - (C1 * dist));
     d = dist * (C1 + d * C2);

     d *= Min((int)pow(2, i), 1024);
  }
  ret = (unsigned int)(d * 1000);


#ifdef _SMP_REQTIMER
  cout << "request: " << "d = " << ret << endl;
#endif

  if (ret < 20)
    ret = 20;
  if (ret > 10000)
    ret = 10000;

  return ret;
}

/*********************************************************************
 Name    : Timer::GetRepairTimer
 Param   :
 Result  :
 Desc    : Erechnet die Laufzeit von Repair Timer
           Parameter enthaelt die Distance zum Sender in Millisekunden
           In i ist spezifiziert, wie oft bereits ein Timer (Request)
           fuer ein bestimmtes Paket gestellt wurde.

 Calledby:
 Uses    :
 *********************************************************************/
unsigned int SMP::Timer::GetRepairTimer(unsigned int Distance, bool LocalRepair) {

  unsigned int ret;
  double  d =  drand48();
  double dist = Distance/1000.; // Dist enthaelt Distanz in Sekunden

#ifdef _SMP_REQTIMER
  cout << " Distance = " << Distance << " drand48() = " << d
       << " SN = " << SN << " toHandle = " << toHandle;
#endif

  if (dist <= 0)
    dist = 0.1;
  
  d = dist * (D1 + d * D2);

  ret = (unsigned int)(d * 1000);

#ifdef _SMP_REQTIMER
  cout << "repair: " << "d = " << ret << endl;
#endif

  if (ret < 10)
    ret = 10;
  if (ret > 10000)
    ret = 10000;
  return ret;
}


/*********************************************************************
 Name    : Timer::Print
 Param   :
 Result  :
 Desc    :
 Calledby:
 Uses    :
 *********************************************************************/
void SMP::Timer::Print() {
  fprintf(SMP::debug, "(");
  switch(ActionId) {
  case RATECTL: fprintf(SMP::debug, "RATE, ");
                break;
  case REQUEST: fprintf(SMP::debug, "REQ, ");
                break;
  case SESSMESS: fprintf(SMP::debug, "SM, ");
                break;
  case REPAIR: fprintf(SMP::debug, "REP, ");
                break;
  case REPEAT: fprintf(SMP::debug, "REPEAT, "); // lj
                break;
  case GIVETHROUGH: fprintf(SMP::debug, "GT, "); // lj
                break;
  }
  fprintf(SMP::debug, "%u, %u, %u, %u, %u, %u\n)", toHandle, SSRC,  SN, StartZeitPunkt,  EndZeitPunkt, GesamtLaufZeit);
}

#endif
