// *******************************************************************
//
//  A Teaching Environement for Remote Lecturing in Higher Education
//
//         Prototype Implementation of a  Scalable Multicast Protocol
//
//                          Werner Geyer
//                      University of Mannheim
//                     Praktische Informatik IV
//                              L 15, 16
//	                    68131 Mannheim
//                             Germany
//
//              geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
//
//  Copyright (c) 1996, 1997
//
//     Lehrstuhl Praktische Informatik IV
//     University of Mannheim
//     Germany
//
//  All rights reserved.
//
//  Permission to use, copy, modify, and distribute this
//  software and its documentation for any purpose and without
//  fee is hereby granted, provided that the above copyright
//  notice appear in all copies and that both that copyright
//  notice and this permission notice appear in supporting
//  documentation.
//
//  We make no representations about the suitability of this
//  software for any purpose.  It is provided "as is" without
//  express or implied warranty.
//
//
//  Author: Werner Geyer
//          Praktische Informatik IV
//          University of Mannheim
//          email: geyer@pi4.informatik.uni-mannheim.de
//
// *******************************************************************
// program : Timer.h
// authors : Martin Grumann, Juergen Vogel
// date    : Aug 1996
// update  : Sep 1997
// desc    : implementation of a Timer
// *******************************************************************

#ifndef __TIMER_H
#define __TIMER_H

#include <math.h>

#ifdef WIN32
#include <sys/timeb.h>
#include <time.h>
#else
#include <sys/time.h>
#endif

#include <stdlib.h>

#include "Member.h"

#ifdef WIN32
extern double drand48();
#endif

double rtcp_intervall(int, double, int, double*, bool);
int Min(int, int);

namespace SMP {
class Timer {
public:

  enum {
    REQUEST     = 0,
    SESSMESS    = 1,
    REPAIR      = 2,
    TERMINATE   = 3,
    RATECTL     = 4,
    REPEAT      = 5,
    GIVETHROUGH = 6,
    DELAYEDSEND = 7,
    DELAYEDSENDU = 8
  };

  Timer(int, unsigned int, unsigned int, int, int, Member*,
        unsigned int, double, bool, signed char, bool, int);
  bool operator!=(const Timer&);
  bool operator==(const Timer&);
  bool operator<(const Timer&);
  bool IsEqual(const Timer&);
  void ReUse(int, unsigned int, int, int, int, Member*,
             unsigned int, double, bool, signed char, bool, int);
  void Print();

  unsigned int GetRequestTimer(unsigned int, int);
  unsigned int GetRepairTimer(unsigned int, bool);

  const double C1;
  const double C2;
  const double D1;
  const double D2;

  int toHandle;
  unsigned int SSRC;
  unsigned int SN;
  int ActionId;
  int Delay;  // Millisekunden
  unsigned int StartZeitPunkt;
  unsigned int GesamtLaufZeit;
  unsigned int EndZeitPunkt;
  Member* Sender;
  unsigned int Erzeuger;
  bool LocalRecovery;
  signed char TTL;
  bool ForeignNack;
  int i;
};
}

#endif
