#include <errno.h>

#ifdef WIN32
#else
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

#include "general.h"
#include <tcl.h>
#include <iostream> // cout


#include "iohelpers.h"

int readn(int fd, void *vptr, int n) {

  int nleft;
  int nread;
  char *ptr;

  ptr = (char*)vptr;
  nleft = n;
  while (nleft > 0) {
    if ((nread = recv(fd, ptr, nleft, 0)) < 0) {
#ifdef WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK) nread = 0;
      else if (WSAGetLastError() == WSAEINTR) nread = 0;
      else return -1;
#else
      if (errno == EINTR) {
	std::cerr << "readn() : interrupted system" << std::endl;
	nread = 0;
      } else if (errno == EWOULDBLOCK) {
	nread = 0;
      }
      else
	return(-1);
#endif
    } else if (nread == 0)
      break;

    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

int recvn(int fd, void *vptr, int n, int flags) {

  int nleft;
  int nread;
  char *ptr;

  ptr = (char*)vptr;
  nleft = n;
  while (nleft > 0) {
    if ((nread = recv(fd, ptr, nleft, flags)) < 0) {
#ifdef WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK) nread = 0;
      else if (WSAGetLastError() == WSAEINTR) nread = 0;
      else return -1;
#else
      if (errno == EINTR) {
	std::cerr << "readn() : interrupted system call" << std::endl;
	nread = 0;
      } else if (errno == EWOULDBLOCK) {
	nread = 0;
      } else
	return(-1);
#endif
    } else if (nread == 0)
      break;

    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

int writen(int fd, const void *vptr, int n) {

  int nleft;
  int nwritten;
  const char *ptr;

  ptr = (char*)vptr;
  nleft = n;
  while(nleft > 0) {

    if ((nwritten = send(fd, ptr, nleft, 0)) <= 0) {
#ifdef WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK) nwritten = 0;
      else if (WSAGetLastError() == WSAEINTR)  nwritten = 0;
      else return -1;
#else
      if (errno == EINTR || errno == EWOULDBLOCK) {
	nwritten = 0;
      } else
	return -1;
#endif
    }
    //cout << "writen " << nwritten << endl;
    nleft -= nwritten;
    ptr += nwritten;
  }

  return(n);
}

int _readn(Tcl_Channel fd, void *vptr, int n) {

  int nleft;
  int nread;
  char *ptr;
  int tclerrno;

  ptr = (char*)vptr;
  nleft = n;
  while (nleft > 0) {
    if ((nread = Tcl_Read(fd, ptr, nleft)) < 0) {
      tclerrno = Tcl_GetErrno();

      if (tclerrno == EINTR) {
	std::cerr << "readn() : interrupted system" << std::endl;
	nread = 0;
      } else if (tclerrno == EWOULDBLOCK) {
	nread = 0;
      }
      else
	return(-1);
    } else if (nread == 0)
      break;

    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

int _writen(Tcl_Channel fd, const void *vptr, int n) {

  int nleft;
  int nwritten;
  char *ptr;
  int tclerrno;

  ptr = (char*)vptr;
  nleft = n;
  while(nleft > 0) {

    
    if ((nwritten = Tcl_Write(fd, ptr, nleft)) <= 0) {
      tclerrno = Tcl_GetErrno();
      if (tclerrno == EINTR || tclerrno == EWOULDBLOCK) {
	nwritten = 0;
      } else
	return -1;
    }
    //cout << "writen " << nwritten << endl;
    nleft -= nwritten;
    ptr += nwritten;
  }

  return(n);
}


int is_multicast(char* saddr) {
  #ifdef DEBUG1
  std::cout << "tools::is_multicast("<<saddr<<")" << endl;
  #endif

  unsigned int ipaddr = inet_addr(saddr);
  if ((ntohl(ipaddr) & 0xE0000000) == 0xE0000000) {
    #ifdef DEBUG1
    std::cout << "tools::is_multicast() " << saddr << " is a multicast address." << std::endl;
    #endif
    return 1;
  } else {
    #ifdef DEBUG1
    std::cout << "tools::is_multicast() " << saddr << " is NOT a multicast address." << std::endl;
    #endif
    return 0;
  };
}

#ifndef WIN32
int closesocket(int sock) {
	return ::close(sock);
}
#endif

